import copy
from collections import ChainMap


class _empty:
    pass


class Context:

    def __init__(self, variables=None, __globals=None, __locals=None, key_server_ctx=None, server_variables=None,
                 vault=None):
        self._global_envs = __globals if __globals is not None else {}  # environment variables global to a deployment
        self._local_envs = __locals if __locals is not None else {}  # environment variables local to a step
        self.env = ChainMap(self._local_envs, self._global_envs)
        self._container = {'vault': vault if vault is not None else {}}

        self._global_variables = variables if variables is not None else {}  # user data shared between steps
        self._key_server_ctx = key_server_ctx
        self._server_variables = server_variables if server_variables is not None else {}
        if key_server_ctx:
            if key_server_ctx not in self._server_variables:
                self._server_variables.update({key_server_ctx: {}})
            self._container['input'] = ChainMap(self._server_variables.get(self._key_server_ctx), self._global_variables)
        else:
            self._container['input'] = self._global_variables

    def __contains__(self, item):
        return item in self.input

    def __getattr__(self, item):
        if item not in self._container:
            raise AttributeError(f"'{item}' is not a valid container")
        return self._container[item]

    def __getitem__(self, item):
        return self.input[item]

    def __setitem__(self, key, value):
        self.input[key] = value
        self.merge_common_variables(key)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def __iter__(self):
        for k, v in self.input.items():
            yield k, v

    def __str__(self):
        return str(self.dict())

    def __repr__(self):
        return repr(self.dict())

    def dict(self):
        return {k: copy.deepcopy(dict(v)) for k, v in self._container.items() if v}

    def get(self, key, default=None):
        return self.input.get(key, default)

    def set(self, item, value):
        self.input.update({item: value})
        self.merge_common_variables(item)

    def pop(self, item):
        return self.input.pop(item)

    def keys(self):
        return self.input.keys()

    def items(self):
        return self.input.items()

    def values(self):
        return self.input.values()

    def local_ctx(self, __locals=None, key_server_ctx=None) -> 'Context':
        return self.__class__(self._global_variables, self._global_envs, __locals,
                              key_server_ctx=key_server_ctx, server_variables=self._server_variables, vault=self.vault)

    def merge_ctx(self, ctx: 'Context'):
        """Function used for loading context from an older orch"""
        # self._global_envs.update(ctx.__global_envs)
        self._local_envs.update(ctx._local_envs)
        self.input.update(ctx.input)
        for server_id in self._server_variables:
            self._server_variables[server_id].update(ctx._server_variables.get(server_id, {}))
        for server_id in ctx._server_variables:
            if server_id not in self._server_variables:
                self._server_variables[server_id] = ctx._server_variables[server_id]

    def merge_common_variables(self, var=_empty):
        common_variables = None
        for server_id in self._server_variables.keys():
            if common_variables is None:
                common_variables = set(self._server_variables[server_id].keys())
            else:
                common_variables = common_variables.intersection(set(self._server_variables[server_id].keys()))

        if common_variables:
            if var in common_variables:
                common_variables = [var]
            for cv in common_variables:
                value = set([self._server_variables[server_id][cv] for server_id in self._server_variables.keys()])
                if len(value) == 1:
                    self._global_variables.update({cv: value.pop()})
                    [self._server_variables[server_id].pop(cv) for server_id in self._server_variables.keys()]



import inspect

import click
from prompt_toolkit import prompt

from dimensigon.dshell.converters import Converter


def prompt_continuation(width, line_number, is_soft_wrap):
    if width >= 2:
        return '.' * (width - 2) + '> '
    elif width == 1:
        return '>'
    else:
        return ''


def code_extension(action_type):
    if action_type == 'SHELL':
        lexer = '.sh'
    elif action_type == 'PYTHON':
        lexer = '.py'
    else:
        lexer = '.txt'
    return lexer


def resolve_callables(kwargs, params):
    prompt_kwargs = dict(kwargs)
    for kk, vv in kwargs.items():
        if callable(vv) and kk not in ('prompt_continuation'):
            sig = inspect.signature(vv)
            callable_kwargs = {kw: params.get(kw, None) for kw in sig.parameters.keys()}
            prompt_kwargs[kk] = vv(**callable_kwargs)
    return prompt_kwargs


def set_parameter(parameter, text, entity, converter: Converter = None):
    if text == '':
        text = None
    elif text == "''" or text == '""':
        text = ""
    else:
        if converter:
            text = converter.load(text)
    if text != entity.get(parameter, None):
        entity[parameter] = text
        return True
    else:
        return False


def prompt_parameter(parameter, entity, form, parent_prompt) -> bool:
    prompt_kwargs = resolve_callables(form[parameter], entity)

    edit = prompt_kwargs.pop('edit', None)
    value = entity.get(parameter, '')
    if 'converter' in prompt_kwargs:
        converter = prompt_kwargs.pop('converter')
        value = converter.dump(value)
    else:
        converter = None
        value = str(value) if value is not None else ''

    if edit:
        text = click.edit(value, extension=edit)
        if text is None:
            text = value
    else:
        text = prompt(
            f"{parent_prompt}.{parameter}>{'>' if prompt_kwargs.get('multiline', False) else ''} ",
            default=value,
            **prompt_kwargs)
    return set_parameter(parameter, text, entity, converter)



from prompt_toolkit.validation import Validator, ValidationError

from dimensigon.dshell import converters as c


class JSON(Validator):

    def validate(self, document):
        text = document.text

        if text:
            try:
                c.JSON.load(text)
            except Exception as e:
                raise ValidationError(message=str(e))


class Bool(Validator):

    def validate(self, document):
        text = document.text

        if text:
            try:
                c.Bool.load(text)
            except:
                raise ValidationError(
                    message=f'Not a valid bool value ({",".join(c.Bool.true)}|{",".join(c.Bool.false)})')


class Int(Validator):

    def validate(self, document):
        text = document.text

        if text:
            try:
                c.Int.load(text)
            except:
                raise ValidationError(message='Not a valid integer')


class Choice(Validator):

    def __init__(self, choices) -> None:
        super().__init__()
        self.choices = choices

    def validate(self, document):
        text = document.text

        if text and text not in self.choices:
            raise ValidationError(message='Invalid choice. Choose from ' + ', '.join(self.choices),
                                  cursor_position=0)


class List(Validator):

    def __init__(self, sep=','):
        self.sep = sep

    def validate(self, document):
        text = document.text

        if text:
            try:
                c.List(self.sep).load(text)
            except Exception as e:
                raise ValidationError(message='Not a valid list. ' + str(e))



from flask import request
from flask_jwt_extended import jwt_required, get_jwt_identity
from flask_restful import Resource
from sqlalchemy import distinct

from dimensigon.domain.entities import Vault
from dimensigon.web import errors, db
from dimensigon.web.decorators import forward_or_dispatch, securizer, lock_catalog, validate_schema
from dimensigon.web.helpers import filter_query, check_param_in_uri
from dimensigon.web.json_schemas import vault_post, vaults_post, vault_put


class VaultList(Resource):

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    def get(self):
        if check_param_in_uri('scopes'):
            return [r[0] for r in
                    db.session.query(distinct(Vault.scope)).filter_by(user_id=get_jwt_identity()).filter_by(
                        deleted=0).all()]
        elif check_param_in_uri('vars'):
            query = db.session.query(distinct(Vault.name)).filter_by(user_id=get_jwt_identity()).filter_by(deleted=0)
            if 'scope' in request.args:
                query = query.filter_by(scope=request.args.get('scope'))
            return [r[0] for r in query.order_by(Vault.name)]
        else:
            query = filter_query(Vault, request.args, exclude=["user_id", "value"]).filter_by(
                user_id=get_jwt_identity())
            return [vault.to_json(no_delete=True, human=check_param_in_uri('human')) for vault in query.all()]

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(vaults_post)
    @lock_catalog
    def post(self):
        data = request.get_json()
        v = Vault.query.get((get_jwt_identity(), data.get('scope', 'global'), data['name']))
        if v:
            raise errors.EntityAlreadyExists("Vault", (data.get('scope', 'global'), data['name']), ("scope", "name"))

        v = Vault(user_id=get_jwt_identity(), scope=data.get('scope', 'global'), name=data['name'], value=data['value'])
        db.session.add(v)
        db.session.commit()
        return {}, 204


class VaultResource(Resource):

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    def get(self, name, scope='global'):
        return Vault.query.get_or_raise((get_jwt_identity(), scope, name)).to_json(human=check_param_in_uri('human'),
                                                                                   no_delete=True)

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(vault_post)
    @lock_catalog
    def post(self, name, scope='global'):
        data = request.get_json()
        v = Vault.query.get_or_raise((get_jwt_identity(), scope, name))

        v.value = data['value']
        db.session.commit()
        return {}, 204

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(vault_put)
    @lock_catalog
    def put(self, name, scope='global'):
        data = request.get_json()
        v = Vault.query.get((get_jwt_identity(), scope, name))
        if v is None:
            v = Vault(user_id=get_jwt_identity(), scope=scope, name=name)
            db.session.add(v)
        v.value = data['value']
        db.session.commit()
        return {}, 204

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @lock_catalog
    def delete(self, name, scope='global'):
        v = Vault.query.get_or_raise((get_jwt_identity(), scope, name))
        v.delete()
        db.session.commit()
        return {}, 204



from flask import request, g
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dimensigon.domain.entities import Server
from dimensigon.web import errors, db
from dimensigon.web.decorators import securizer, forward_or_dispatch, lock_catalog, validate_schema
from dimensigon.web.helpers import filter_query, check_param_in_uri
from dimensigon.web.json_schemas import server_patch, servers_delete


class ServerList(Resource):

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    def get(self):
        query = filter_query(Server, request.args)
        return [s.to_json(add_gates=check_param_in_uri('gates'),
                          human=check_param_in_uri('human'),
                          no_delete=True,
                          add_ignore=True) for s in
                query.all()]

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(servers_delete)
    @lock_catalog
    def delete(self):
        servers = [Server.query.get_or_raise(s_id) for s_id in request.get_json()['server_ids']]
        for server in servers:
            if server == g.server:
                raise errors.ServerDeleteError
            # remove associated routes
            if server.route:
                db.session.delete(server.route)
            server.delete()
        db.session.commit()

        return {}, 204


class ServerResource(Resource):
    @forward_or_dispatch()
    @jwt_required()
    @securizer
    def get(self, server_id):
        return Server.query.get_or_raise(server_id).to_json(add_gates=check_param_in_uri('gates'),
                                                            human=check_param_in_uri('human'),
                                                            no_delete=True,
                                                            add_ignore=True)

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @validate_schema(server_patch)
    @lock_catalog
    def patch(self, server_id):
        json_data = request.get_json()
        resp_data = {}

        server = Server.query.get_or_raise(server_id)
        if 'all' in json_data.get('granules', []):
            raise errors.KeywordReserved("'all' is a reserved granule")

        new_granules = set(json_data.get('granules', [])) - set(server.granules)
        if new_granules:
            server.granules.extend(set(new_granules) - set(server.granules))

        new_gate_ids = []
        for gate in json_data.get('gates'):
            g = server.add_new_gate(gate['dns_or_ip'], gate['port'], gate.get('hidden'))
            new_gate_ids.append(g.id)
        if new_gate_ids:
            resp_data.update(gate_ids=new_gate_ids)

        if 'ignore_on_lock' in json_data:
            server.l_ignore_on_lock = json_data.get('ignore_on_lock')

        db.session.commit()

        if resp_data:
            return resp_data, 200
        else:
            return {}, 204

    @forward_or_dispatch()
    @jwt_required()
    @securizer
    @lock_catalog
    def delete(self, server_id):
        server = Server.query.get_or_raise(server_id)
        if server == g.server:
            raise errors.ServerDeleteError
        # remove associated routes
        db.session.delete(server.route)
        server.delete()
        db.session.commit()

        return {}, 204



import base64
import concurrent
import copy
import datetime
import functools
import json
import logging
import pickle
import re
import threading
import time
import typing as t
import uuid
from abc import ABC, abstractmethod
from collections import OrderedDict
from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import contextmanager
from functools import partial

import jsonschema
from flask import current_app, has_app_context, g
from flask_jwt_extended import create_access_token
from sqlalchemy.orm import sessionmaker

from dimensigon import defaults
from dimensigon.domain.entities import Server, Orchestration, Step, StepExecution, OrchExecution, User, Scope
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.use_cases import lock as lock
from dimensigon.use_cases.lock import locker_scope_enabled
from dimensigon.use_cases.operations import CompletedProcess, IOperationEncapsulation, create_operation
from dimensigon.utils.dag import DAG
from dimensigon.utils.event_handler import Event
from dimensigon.utils.helpers import get_now, format_exception
from dimensigon.utils.typos import Id
from dimensigon.utils.var_context import Context
from dimensigon.web import db, errors, executor, DimensigonFlask
from dimensigon.web.network import post

# if t.TYPE_CHECKING:


logger = logging.getLogger('dm.deployment')


def exec_safe(code, locals=None):
    # TODO: redefine builtin scope
    # byte_code = compile_restricted(code, '<string>', 'exec')
    # safe_builtins.update(json=json)
    # safe_builtins.update(yaml=yaml)
    # safe_builtins.update(re=re)
    # exec(byte_code, {'__builtins__': safe_builtins,
    #                  '_write_': full_write_guard,
    #                  '_getiter_': default_guarded_getiter},
    #      locals)
    exec(code, {}, locals)


class ICommand(ABC):
    _id: Id

    @property
    @abstractmethod
    def success(self) -> bool:
        """
        Property that returns True if the command executed successfully
        Returns
        -------

        """

    @property
    @abstractmethod
    def result(self) -> t.Dict[Id, CompletedProcess]:
        """
        Property to get all the results

        Returns
        -------
        dict with all the results
        """

    @abstractmethod
    def invoke(self, timeout=None) -> bool:
        """
        Implement the do process. If the process ends as expected it must set _success to true.

        Returns
        -------
        bool:
            True if the do process ends as expected. False otherwise
        """

    @abstractmethod
    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Method to implement the undo process. Executed only if success is true

        Returns
        -------
        bool:
            True if the undo process ends up OK. False otherwise
        """

    @property
    def id(self) -> Id:
        """
        Property that returns the ID of the command
        Returns
        -------

        """
        return self._id

    def __repr__(self):
        return f"{self.__class__.__name__} {self.id}"


def elapsed_times(tag):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            if not hasattr(self, '_elapsed_times'):
                self._elapsed_times = {}
            start = time.time()
            ret = f(self, *args, **kwargs)
            self._elapsed_times[tag] = time.time() - start
            return ret

        return wrapper

    return decorator


def extract_container_var(var):
    if var.startswith('input.'):
        source = var.split('.', 1)[1]
        container_name = 'input'
    elif '.' not in var:
        source = var
        container_name = 'input'
    else:
        container_name, source = var.split('.', 1)
    return container_name, source


def normalize_container_var(var):
    return '.'.join(extract_container_var(var))



reserved_words = ['mapping', 'required', 'output']


class ImplementationCommand(ICommand):

    def __init__(self, implementation: IOperationEncapsulation,
                 var_context: Context = None,
                 id_=None,
                 register: 'RegisterStepExecution' = None,
                 pre_process=None,
                 post_process=None,
                 signature=None
                 ):
        self._id = id_ or id(self)
        self.implementation = implementation
        self.var_context = var_context
        self.params = DictDict()
        self.register = register
        self.step_execution_id = None
        self.pre_process_code = pre_process
        self.post_process_code = post_process
        self.signature = signature or {}
        self._elapsed_times = {}
        self._cp: CompletedProcess = None

    def create_step_execution(self):
        if self.register:
            step_execution_id = self.register.create_step_execution(self)
            self.var_context.env.update(step_execution_id=step_execution_id)
            self.step_execution_id = step_execution_id

    def register_execution(self):
        if self.register:
            self.register.save_step_execution(self, params=self.params, **self._elapsed_times)

    @elapsed_times('pre_process_time')
    def pre_process(self):
        if self.pre_process_code:
            try:
                local = dict(vc=self.var_context, cp=self._cp)
                exec_safe(self.pre_process_code, local)
            except Exception as e:
                self._cp = CompletedProcess(success=False, stderr=f"Pre-Process error: {format_exception(e)}")

    def extract_params(self):
        """
        Turns over only params from var context to params
        :return:
        """
        if not self._cp:
            try:
                if self.signature:
                    # get required variables
                    required = DictSet()
                    for r in self.signature.get('required', []):
                        container_name, var = extract_container_var(r)
                        required[container_name].add(var)
                        # del container_name
                    # containers in signature
                    container_names = [k for k in self.signature.keys() if k not in reserved_words]
                    # add containers to turn over params from required
                    container_names = set(container_names) | set(required.keys())
                    container_names -= {'env'}  # remove env container cause is going to be passed explicitly

                    # resolve mapping values
                    for dest, value in self.signature.get('mapping', {}).items():
                        if isinstance(value, dict) and len(value) == 1 and 'from' in value:
                            action, source = tuple(value.items())[0]
                            container_name, var = extract_container_var(source)
                            try:
                                self.params['input'].update({dest: getattr(self.var_context, container_name, {})[var]})
                            except KeyError:
                                if dest in required.get('input', []):
                                    se = StepExecution.query.get(self.step_execution_id)
                                    raise errors.MissingParameters([source],
                                                                   se.step, se.server)
                        else:
                            self.params['input'][dest] = value

                    schema2validate = {'type': 'object', 'properties': {}}
                    for container_name in container_names:
                        schema2validate['properties'].update(
                            {container_name: dict(type="object", properties=self.signature.get(container_name, {}),
                                                  required=list(required[container_name]))})
                        for var, value in self.signature.get(container_name, {}).items():
                            if var not in self.params[container_name]:
                                try:
                                    self.params[container_name].update(
                                        {var: getattr(self.var_context, container_name, {})[var]})
                                except KeyError:
                                    if 'default' in value:
                                        self.params[container_name][var] = value['default']
                                    elif var in required.get(container_name, []):
                                        if has_app_context():
                                            raise errors.MissingParameters([f"{container_name}.{var}"],
                                                                           Step.query.get(self.id[1]),
                                                                           Server.query.get(self.id[0]))
                                        else:
                                            raise errors.MissingParameters([f"{container_name}.{var}"])
                        # add variables specified in required
                        for var in required.get(container_name, []):
                            if var not in self.params[container_name]:
                                try:
                                    self.params[container_name].update(
                                        {var: getattr(self.var_context, container_name, {})[var]})
                                except KeyError:
                                    raise errors.MissingParameters([f"{container_name}.{var}"],
                                                                   Step.query.get(self.id[1]),
                                                                   Server.query.get(self.id[0]))

                    if schema2validate:
                        jsonschema.validate(self.params, schema2validate)
                else:
                    self.params = self.var_context.dict()
            except KeyError as e:
                self._cp = CompletedProcess(success=False, stderr=f"Variable {e} not found")
            except jsonschema.ValidationError as e:
                self._cp = CompletedProcess(success=False, stderr=f"Param validation error: {e}")
            except errors.MissingParameters as e:
                self._cp = CompletedProcess(success=False, stderr=f"{e}")
            except Exception as e:
                self._cp = CompletedProcess(success=False, stderr=f"{format_exception(e)}")

    def _invoke(self, timeout):
        if not self._cp:
            try:
                self._cp = self.implementation.execute(self.params, timeout=timeout, context=self.var_context)
            except Exception as e:
                self._cp = CompletedProcess(success=False, stderr=f"Execution error: {e}")
                logger.exception(f"Exception on execution {self.id}")

    @elapsed_times('post_process_time')
    def post_process(self):
        if self.post_process_code:
            try:
                local = dict(vc=self.var_context, cp=self._cp)
                exec_safe(self.post_process_code, local)
            except Exception as e:
                self._cp.success = False
                if self._cp.stderr:
                    self._cp.stderr = f"{self._cp.stderr}\nPost-Process error: {format_exception(e)}"
                else:
                    self._cp.stderr = f"Post-Process error: {format_exception(e)}"

    def check_output_variables(self):
        if self._cp.success and self.signature.get('output', []):
            output = set(self.signature.get('output', []))
            missing = output - set(self.var_context.keys())
            if missing:
                self._cp.success = False
                if self._cp.stderr:
                    self._cp.stderr = f"{self._cp.stderr}\nMissing output values: {', '.join(missing)}"
                else:
                    self._cp.stderr = f"Missing output values: {', '.join(sorted(missing))}"

    def invoke(self, timeout=None) -> t.Optional[bool]:
        if not self._cp:
            self.create_step_execution()

            self.pre_process()

            self.extract_params()

            self._invoke(timeout)

            self.post_process()

            self.check_output_variables()

            self.register_execution()

        return self.success

    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        return {self.id: self._cp}

    @property
    def success(self) -> t.Optional[bool]:
        return getattr(self._cp, 'success', None)


class UndoCommand(ImplementationCommand):

    def __init__(self, *args, stop_on_error: bool = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop_on_error = stop_on_error

    def undo(self, timeout=None) -> t.Optional[bool]:
        return True

    def __iter__(self):
        return [self]


class Command(ImplementationCommand):

    def __init__(self, *args,
                 undo_command: t.Union['CompositeCommand', UndoCommand] = None,
                 stop_on_error: bool = None, stop_undo_on_error: bool = None, undo_on_error: bool = True, **kwargs
                 ):
        """
        
        Parameters
        ----------
        implementation:
            operation to perform.
        undo_command:
            undo command
        var_context:
            var_context used to execute the implementation
        undo_on_error:
            sets whether to execute "undo" function when "invoke" terminated incorrectly
        """
        super().__init__(*args, **kwargs)
        self.undo_command = undo_command
        self.stop_on_error = stop_on_error
        self.stop_undo_on_error = stop_undo_on_error
        self.undo_on_error = undo_on_error
        self._cp: t.Optional[CompletedProcess] = None

    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        e = {}
        e.update({self.id: self._cp})
        if self.undo_command and self.undo_command.success is not None:
            e.update(self.undo_command.result)
        return e

    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Executes undo operation

        Returns
        -------
        bool:
            True if all undo commands that run ended up successfully
            False if any undo command ended up badly
            None if undo operation not executed
        """
        if self.undo_command:
            if (self.success is True or (
                    self.success is False and self.undo_on_error)) and self.undo_command.success is None:
                return self.undo_command.invoke(timeout=timeout)
            else:
                return self.undo_command.success
        else:
            return True

    def __iter__(self):
        return [self]


class CompositeCommand(ICommand):

    def __init__(self, dict_tree: t.Dict[ICommand, t.List[ICommand]],
                 stop_on_error: bool, stop_undo_on_error: bool = None,
                 id_=None, executor: concurrent.futures.Executor = None,
                 register: 'RegisterStepExecution' = None, var_context: Context = None):
        """

        Parameters
        ----------
        dict_tree:
            dict tree containing command nodes as keys and it's neighbours as a list of commands. See examples for more
            information
        undo_on_error:
            executes the undo process even if the command ended up with error
        stop_on_error:
            executes all the commands regardless if a command succeed or fail
        stop_undo_on_error:
            executes all the undo commands regardless if an undo command succeed or fail
        id_:
            command identifier
        executor:
            async call executor. defaults to concurrent.futures.process.ProcessPoolExecutor
        """
        self._id = id_
        self._dag = DAG().from_dict_of_lists(dict_tree)
        self.stop_on_error = stop_on_error
        self.stop_undo_on_error = stop_undo_on_error
        self.executor = executor or ThreadPoolExecutor(max_workers=4)
        self.register = register
        self.var_context = var_context

    @property
    def success(self) -> t.Optional[bool]:
        success_list = [n.success for n in self._dag]
        if success_list == [None] * len(success_list):
            return None
        elif success_list == [True] * len(success_list):
            return True
        else:
            return False

    def invoke(self, timeout=None) -> bool:
        # logging.basicConfig(format='%(threadName)s %(levelname)-8s %(message)s')
        # logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
        res = []
        stop = False
        level = 1
        start = time.time()
        while level <= self._dag.depth:
            commands = self._dag.get_nodes_at_level(level)
            duration = time.time() - start
            left = max(timeout - duration, 0) if timeout else None
            if len(commands) == 1:
                r = commands[0].invoke(timeout=left)
                res.append(r)
                stop_on_error = commands[0].stop_on_error if commands[0].stop_on_error is not None \
                    else self.stop_on_error
                if r is False and stop_on_error is True:
                    break
            else:
                futures = {}

                for command in commands:
                    futures.update(
                        {command.id: self.executor.submit(command.invoke, timeout=left)})

                stop = None
                while len(futures) > 0 and (left is None or (time.time() - start) < left) and stop is not True:
                    for cmd_id, future in dict(futures).items():
                        r = None
                        try:
                            r = future.result(timeout=0.5)
                        except concurrent.futures.TimeoutError:
                            pass
                        except Exception as e:
                            logger.exception(f"Error while executing step {cmd_id}")
                            del futures[cmd_id]
                            r = False
                        else:
                            del futures[cmd_id]
                        if r is not None:
                            res.append(r)
                            cmd = list(filter(lambda c: c.id == cmd_id, commands))[0]
                            stop_on_error = cmd.stop_on_error if cmd.stop_on_error is not None else self.stop_on_error
                            if r is False and stop_on_error is True:
                                stop = True

                    duration = time.time() - start
                    left = max(timeout - duration, 0) if timeout else None
                # if self.var_context:
                #     self.var_context.merge_common_variables()
                if stop is True or (timeout is not None and (time.time() - start) > timeout):
                    break
            level += 1
        return all(res) if len(res) > 0 else None

    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Executes undo operation

        Returns
        -------
        bool:
            True if all undo commands that run ended up successfully
            False if any undo command ended up badly
            None if undo operation not executed
        """
        if self.stop_undo_on_error is None:
            raise RuntimeError(f'stop_undo_on_error not set for command {self.id}')
        res = []
        level = self._dag.depth
        start = time.time()
        duration = time.time() - start
        left = max(timeout - duration, 0) if timeout else None
        while level > 0 and (left is None or (time.time() - start) < left):
            commands = self._dag.get_nodes_at_level(level)
            duration = time.time() - start
            left = max(timeout - duration, 0) if timeout else None
            if len(commands) == 1:
                r = commands[0].undo(timeout=left)
                if r is not None:
                    res.append(r)
                stop_undo_on_error = commands[0].stop_undo_on_error if commands[0].stop_undo_on_error is not None \
                    else self.stop_undo_on_error
                if r is False and stop_undo_on_error is True:
                    break
            else:
                futures = {}

                for command in commands:
                    futures.update(
                        {command.id: self.executor.submit(command.undo, timeout=left)})

                stop = None
                while len(futures) > 0 and (left is None or (time.time() - start) < left):
                    for cmd_id, future in dict(futures).items():
                        r = None
                        try:
                            r = future.result(timeout=0.5)
                        except concurrent.futures.TimeoutError:
                            pass
                        except Exception as e:
                            logger.exception(f"Error while executing step {cmd_id}")
                            del futures[cmd_id]
                            r = False
                        else:
                            del futures[cmd_id]
                        if r is not None:
                            res.append(r)
                            cmd = list(filter(lambda c: c.id == cmd_id, commands))[0]
                            stop_undo_on_error = cmd.stop_undo_on_error if cmd.stop_undo_on_error is not None \
                                else self.stop_undo_on_error
                            if r is False and cmd.stop_undo_on_error is True:
                                stop = True
                    duration = time.time() - start
                    left = max(timeout - duration, 0) if timeout else None

                if stop is True or (timeout is not None and (time.time() - start) > timeout):
                    break
            level -= 1
        return all(res) if len(res) > 0 else None

    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        e = {}
        for n in self._dag.nodes:
            if n.success is not None:
                e.update(n.result)
        return e

    def __len__(self):
        return len(self._dag)

    def __iter__(self):
        return self._dag


class ProxyImplementation:
    _command: ImplementationCommand

    def __init__(self, server: Id):
        """
        Parameters
        ----------
        server:
            Server Id to execute the command
        implementation:
            operation to perform.
        undo_implementation:
            undo operation to perform
        var_context:
            var_context to pass to the execution
        undo_on_error:
            if the invoke ended up with an error, the undo process will be executed if undo_on_error is True
        id_:
            command identifier
        timeout:
            timeout when waiting response from remote server when invoke and undo executed
        """
        self.__dict__['_server'] = server
        self.__dict__['_app'] = current_app._get_current_object()
        self.__dict__['_auth'] = None
        self.__dict__['_completion_event'] = threading.Event()
        self.__dict__['_command'] = None

    #
    # proxying (special cases)
    #

    def __getattr__(self, item):
        return getattr(object.__getattribute__(self, "_command"), item)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_command"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_command"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_command"))

    def __str__(self):
        return str(object.__getattribute__(self, "_command"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_command"))

    def __hash__(self):
        return hash(object.__getattribute__(self, "_command"))

    @property
    def server(self) -> 'Server':
        return self._server

    def callback_completion_event(self, event: Event):
        """callback executed on response to the invoke command on remote server
        """
        if 'step_execution' in event.data:
            step_exec_json = event.data.get('step_execution')
            self._command._cp = CompletedProcess(success=step_exec_json.get('success'),
                                                 stdout=step_exec_json.get('stdout'),
                                                 stderr=step_exec_json.get('stderr'),
                                                 rc=step_exec_json.get('rc'),
                                                 start_time=datetime.datetime.strptime(step_exec_json.get('start_time'),
                                                                                       defaults.DATETIME_FORMAT) if step_exec_json.get(
                                                     'start_time') else None,
                                                 end_time=datetime.datetime.strptime(step_exec_json.get('end_time'),
                                                                                     defaults.DATETIME_FORMAT) if step_exec_json.get(
                                                     'end_time') else None)
        else:
            self._command._cp = CompletedProcess(success=False, stdout=str(event.data),
                                                 stderr=f'Unknown message got on completion event.')

        self._completion_event.set()

    def _invoke(self, timeout) -> bool:
        """
        invokes the command on the remote server

        Returns
        -------
        bool:
            True if command ended up gracefully. False otherwise

        Raises
        ------
        TimeoutError:
            raised when timeout reached while waiting the response back from the remote server
        """
        ctx = None
        if not has_app_context():
            ctx = self._app.app_context()
            ctx.push()
        try:
            self.__dict__['_server'] = Server.query.get(self._server)

            # set a timeout if none to avoid infinite wait in event
            if timeout is None:
                timeout = defaults.TIMEOUT_REMOTE_COMMAND
            if not self._command._cp:
                auth = HTTPBearerAuth(
                    create_access_token(self._command.var_context.env['executor_id'], datetime.timedelta(seconds=15)))
                start = time.time()
                data = dict(operation=base64.b64encode(pickle.dumps(self._command.implementation)).decode('ascii'),
                            var_context=base64.b64encode(pickle.dumps(self._command.var_context)).decode('ascii'),
                            params=base64.b64encode(pickle.dumps(self._command.params)).decode('ascii'),
                            timeout=timeout,
                            step_id=str(self.id[1]),
                            orch_execution=self._command.register.json_orch_execution,
                            event_id=str(uuid.uuid4()))
                resp = post(server=self.server, view_or_url='api_1_0.launch_operation', json=data, auth=auth,
                            timeout=timeout)
                if resp.code == 204:
                    current_app.events.register(data['event_id'], self.callback_completion_event)
                    event = self._completion_event.wait(timeout=timeout - (time.time() - start))
                    if event is not True:
                        self._command._cp = CompletedProcess(success=False, stdout='',
                                                             stderr=f'Timeout of {timeout} reached waiting '
                                                                    f'server operation completion')

                elif resp.code == 200:
                    self.callback_completion_event(Event(None, data=resp.msg))
                elif resp.code:
                    if isinstance(resp.msg, dict):
                        msg = json.dumps(resp.msg)
                    else:
                        msg = str(resp.msg)

                    self._command._cp = CompletedProcess(success=False, stdout='',
                                                         stderr=msg, rc=resp.code)

        finally:
            if ctx:
                ctx.pop()
        return self.success


class ProxyCommand(ProxyImplementation, Command):

    def __init__(self, server_id: Id, *args, **kwargs):
        super().__init__(server_id)
        self.__dict__['_command'] = Command(*args, **kwargs)


class ProxyUndoCommand(ProxyImplementation, UndoCommand):

    def __init__(self, server_id: Id, *args, **kwargs):
        super().__init__(server_id)
        self.__dict__['_command'] = UndoCommand(*args, **kwargs)


def _create_server_undo_command(executor, current_server, server_id: Id, s: Step, var_context: Context, register,
                                d=None, s2cc=None) -> t.Optional[t.Union[UndoCommand, CompositeCommand]]:
    def iterate_tree(_cls, _step: Step, _d, _s2cc):
        if _step in _s2cc:
            _uc = _s2cc[_step]
        else:
            stop_on_error = _step.step_stop_on_error if _step.step_stop_on_error is not None else s.stop_undo_on_error
            _uc = _cls(create_operation(_step),
                       id_=(str(server_id), str(_step.id)),
                       var_context=var_context.local_ctx({'server_id': server_id}, key_server_ctx=server_id),
                       pre_process=_step.pre_process,
                       post_process=_step.post_process,
                       stop_on_error=stop_on_error,
                       register=register,
                       signature=_step.schema)
            _s2cc[_step] = _uc
        if _uc not in _d:
            _d[_uc] = []
        for child_step in _step.children_undo_steps:
            _d[_uc].append(iterate_tree(_cls, child_step, _d, _s2cc))
        return _uc

    d = d or {}
    s2cc = s2cc or {}
    if str(server_id) == str(current_server.id):
        u_cmd_cls = UndoCommand
    else:
        u_cmd_cls = partial(ProxyUndoCommand, server_id)
    uc = None
    for child_undo_step in s.children_undo_steps:
        uc = iterate_tree(u_cmd_cls, child_undo_step, d, s2cc)
    if len(d) == 1:
        return uc
    elif len(d) > 1:
        return CompositeCommand(dict_tree=d, stop_on_error=s.stop_undo_on_error,
                                stop_undo_on_error=None,
                                id_=('undo', str(s.id)),
                                executor=executor)


def create_cmd_from_orchestration(orchestration: Orchestration, var_context: Context,
                                  hosts: t.Dict[str, t.List[Id]],
                                  executor: concurrent.futures.Executor,
                                  register: 'RegisterStepExecution') -> CompositeCommand:
    current_server = Server.get_current()

    def create_do_cmd_from_step(_step: Step, _s2cc):
        d = {}
        if _step in _s2cc:
            cc = _s2cc[_step]
        else:
            activate_server_ctx = len([server_id for target in _step.target for server_id in hosts[target]]) > 1
            for target in _step.target:
                for server_id in hosts[target]:
                    if str(server_id) == str(current_server.id):
                        cls = Command
                    else:
                        cls = partial(ProxyCommand, server_id)

                    var_kwargs = dict()
                    if activate_server_ctx:
                        var_kwargs.update(key_server_ctx=server_id)
                    c = cls(create_operation(_step),
                            undo_command=_create_server_undo_command(executor, current_server, server_id,
                                                                     _step, var_context, register),
                            var_context=var_context.local_ctx({'server_id': server_id}, **var_kwargs),
                            stop_on_error=_step.stop_on_error,
                            stop_undo_on_error=None,
                            undo_on_error=_step.undo_on_error,
                            pre_process=_step.pre_process,
                            post_process=_step.post_process,
                            register=register,
                            id_=(str(server_id), str(_step.id)),
                            signature=_step.schema)

                    d[c] = []
            if len(d) == 1:
                cc = c
            elif len(d) > 1:
                cc = CompositeCommand(dict_tree=d,
                                      stop_on_error=False,
                                      stop_undo_on_error=False,
                                      id_=str(_step.id), executor=executor, register=register, var_context=var_context)
            else:
                cc = None
        return cc

    def iterate_tree(_step: Step, _d, _s2cc):
        if _step in s2cc_map:
            _c = _s2cc[_step]
        else:
            _c = create_do_cmd_from_step(_step, _s2cc)
            _s2cc[_step] = _c
            if _c not in _d:
                _d[_c] = []
            for child_step in _step.children_do_steps:
                _d[_c].append(iterate_tree(child_step, _d, _s2cc))
        return _c

    root_steps = orchestration.root
    tree = {}
    s2cc_map = {}
    for step in root_steps:
        iterate_tree(step, tree, s2cc_map)

    return CompositeCommand(dict_tree=tree,
                            stop_undo_on_error=orchestration.stop_undo_on_error,
                            stop_on_error=orchestration.stop_on_error,
                            id_=str(orchestration.id), executor=executor, register=register)


ORCH_EXEC_PATTERN = re.compile(r"^orch_execution_id=([a-zA-Z0-9-]{36})$")


class RegisterStepExecution:
    def __init__(self, orch_execution: OrchExecution, app: DimensigonFlask = None):
        if app:
            self.app = app
        else:
            self.app = current_app._get_current_object()

        self.json_orch_execution = orch_execution.to_json()  # json_orch_execution property is used when running ProxyCommand
        self._store: t.Dict[Id, StepExecution] = OrderedDict()
        if hasattr(app, 'dm') and getattr(app.dm, 'engine'):
            engine = app.dm.engine
        else:
            engine = db.get_engine()
        self.Session = sessionmaker(bind=engine, autoflush=False)

        # def receive_after_transaction_create(session: Session, transaction:SessionTransaction):
        #     print(f"Created transaction {transaction}{transaction._state}({transaction.parent}) on session {session}()")
        #
        # def receive_after_transaction_end(session, transaction):
        #     print(f"Ended transaction {transaction}{transaction._state}({transaction.parent}) on session {session}")
        #
        # from sqlalchemy import event
        # event.listen(self.Session, 'after_transaction_create', receive_after_transaction_create)
        # event.listen(self.Session, 'after_transaction_end', receive_after_transaction_end)

    @property
    def id(self):
        return self.json_orch_execution.get('id')

    @contextmanager
    def session_scope(self):
        """Provide a transactional scope around a series of operations."""
        session = self.Session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            logger.exception("Unable to commit")
        finally:
            session.close()

    def update_orch_execution(self, **kwargs):
        with self.session_scope() as s:
            orch_execution = s.query(OrchExecution).get(self.json_orch_execution.get('id'))
            for key, value in kwargs.items():
                setattr(orch_execution, key, value)
            self.json_orch_execution = orch_execution.to_json()

    def create_step_execution(self, command):
        ident = str(uuid.uuid4())
        with self.session_scope() as s:
            se = StepExecution(id=ident, step_id=command.id[1],
                               server_id=command.id[0], orch_execution_id=self.json_orch_execution.get('id'),
                               start_time=get_now())
            s.add(se)
        self._store[ident] = se
        return ident

    def save_step_execution(self, command: ImplementationCommand, params=None, pre_process_time=None,
                            execution_time=None,
                            post_process_time=None):
        with self.session_scope() as s:
            se = s.merge(self._store[command.step_execution_id])
            se.load_completed_result(command._cp)
            se.params = params
            se.pre_process_elapsed_time = pre_process_time
            se.execution_elapsed_time = execution_time
            se.post_process_elapsed_time = post_process_time
            se.end_time = get_now()
            if command._cp.stdout and ORCH_EXEC_PATTERN.match(command._cp.stdout):
                se.child_orch_execution_id = ORCH_EXEC_PATTERN.match(command._cp.stdout)[1]

    def commit_data(self):
        pass
        # if self._store:
        #     with self.session_scope() as s:
        #         for step in self._store.values():
        #             s.add(step)
        #         try:
        #             s.commit()
        #         except:
        #             s.rollback()
        #             logger.exception("Unable to commit data")
        #         else:
        #             self._store.clear()


class DictSet(dict):

    def __getitem__(self, item):
        if item not in self:
            super().__setitem__(item, set())
        return super().__getitem__(item)

    def __sub__(self, other):
        missing = []
        for c, v in self.items():
            m = v - other[c]
            missing.extend([f"{c}.{vv}" for vv in m])
        return missing


class DictDict(dict):

    def __getitem__(self, item):
        if item not in self:
            super().__setitem__(item, {})
        return super().__getitem__(item)

    def __sub__(self, other):
        missing = []
        for c, v in self.items():
            m = v - other[c]
            missing.extend([f"{c}.{vv}" for vv in m])
        return missing


def validate_input_chain(validatable: t.Union[Orchestration, Step], params: t.Dict[str, set]) -> \
        t.Dict[Step, t.Set[str]]:
    """validates against the validatable that params have all needed parameters

    Parameters
    ----------
    validatable: object that carries the schema
    params:      dict with variables grouped by container

    Returns
    -------
    returns a dict like object with all missed parameters per step
    """

    def split_container(_var):
        if '.' in var:
            c_name, v_name = var.split('.', 1)
        else:
            c_name, v_name = 'input', var
        return c_name, v_name

    iterable = validatable.root if isinstance(validatable, Orchestration) else validatable.children
    not_found = {}
    if not isinstance(params, DictSet):
        params = DictSet(params)
    current_params = DictSet(copy.deepcopy(params))
    for step in iterable:
        p = set(step.schema.get('required', []))
        required_params = DictSet()
        for var in p:
            if '.' in var:
                container_name, var_name = var.split('.', 1)
            else:
                container_name, var_name = 'input', var
            required_params[container_name].add(var_name)

        m = step.schema.get('mapping', {})
        if m:
            for dest, value in m.items():
                if isinstance(value, dict) and len(value) == 1 and 'from' in value:
                    action, source = tuple(value.items())[0]
                    if dest in required_params['input']:
                        current_params['input'].add(dest)
                        # add source as required
                        container_name, var_name = split_container(source)
                        required_params[container_name].add(var_name)
                else:
                    current_params['input'].add(dest)

        # add default values into current_params
        for k, v in step.schema.get('input', {}).items():
            if 'default' in v:
                current_params['input'].add(k)

        missing = required_params - current_params
        if missing:
            raise errors.MissingParameters(list(missing), step)

        out_params = set(step.schema.get('output', []))
        params['input'].update(out_params)

        r_not_found = validate_input_chain(step, params=params)
        not_found.update(r_not_found)
    return not_found


def deploy_orchestration(orchestration: t.Union[Id, Orchestration],
                         hosts: t.Dict[str, t.Union[t.List[Id]]],
                         var_context: 'Context' = None,
                         execution: t.Union[Id, OrchExecution] = None,
                         executor: t.Union[Id, User] = None,
                         execution_server: t.Union[Id, Server] = None,
                         lock_retries=2,
                         lock_delay=3, timeout=None) -> Id:
    """deploy the orchestration

    Args:
        orchestration: id or orchestration to execute
        hosts: Mapping to all distributions
        var_context: Context configuration
        execution: id or execution to associate with the orchestration. If none, a new one is created
        executor: id or User who executes the orchestration
        execution_server: id or User who executes the orchestration
        lock_retries: tries to lock for orchestration N times
        lock_delay: delay between retries
    Returns:
        OrchExecution ID

    Raises:
        Exception: if anything goes wrong
    """
    execution = execution or var_context.env.get('orch_execution_id')
    executor = executor or var_context.env.get('executor_id')
    hosts = hosts or var_context.get('hosts')
    if not isinstance(orchestration, Orchestration):
        orchestration = db.session.query(Orchestration).get(orchestration)
    if not isinstance(execution, OrchExecution):
        exe = None
        if execution is not None:
            exe = db.session.query(OrchExecution).get(execution)
        if exe is None:
            if not isinstance(executor, User):
                executor = db.session.query(User).get(executor)
            if executor is None:
                raise ValueError('executor must be set')
            if not isinstance(execution_server, Server):
                if execution_server is None:
                    try:
                        execution_server = g.server
                    except AttributeError:
                        execution_server = Server.get_current()
                    if execution_server is None:
                        raise ValueError('execution server not found')
                else:
                    execution_server = db.session.query(Server).get(execution_server)
            exe = OrchExecution(id=execution, orchestration_id=orchestration.id, target=hosts,
                                params=dict(var_context),
                                executor_id=executor.id, server_id=execution_server.id)
            db.session.add(exe)
            db.session.commit()

    else:
        exe = execution
    current_app.logger.debug(
        f"Execution {exe.id}: Launching orchestration {orchestration} on {hosts} with {var_context}")

    return _deploy_orchestration(orchestration, var_context, hosts, exe, lock_retries, lock_delay, timeout)


def _deploy_orchestration(orchestration: Orchestration,
                          var_context: 'Context',
                          hosts: t.Dict[str, t.List[Id]],
                          execution: OrchExecution,
                          lock_retries,
                          lock_delay,
                          timeout
                          ) -> Id:
    """
    Parameters
    ----------
    orchestration
        orchestration to deploy
    params
        parameters to pass to the steps

    Returns
    -------
    t.Tuple[bool, bool, t.Dict[int, dpl.CompletedProcess]]:
        tuple with 3 values. (boolean indicating if invoke process ended up successfully,
        boolean indicating if undo process ended up successfully,
        dict with all the executions). If undo process not executed, boolean set to None
    """
    rse = RegisterStepExecution(execution)
    kwargs = dict()
    kwargs['start_time'] = execution.start_time or get_now()
    cc = create_cmd_from_orchestration(orchestration, var_context, hosts=hosts, register=rse, executor=executor)

    # convert UUID into str as in_ filter does not handle UUID type
    all = [str(s) for s in hosts['all']]
    servers = Server.query.filter(Server.id.in_(all)).all()
    scope_enabled = locker_scope_enabled(Scope.ORCHESTRATION)
    if scope_enabled:
        try:
            applicant = lock.lock(Scope.ORCHESTRATION, servers, applicant=var_context.env.get('root_orch_execution_id'),
                                  retries=lock_retries, delay=lock_delay)
        except errors.LockError as e:
            kwargs.update(success=False, message=str(e))
            rse.update_orch_execution(**kwargs)
            raise
    try:
        kwargs['success'] = cc.invoke(timeout=timeout)
        if not kwargs['success'] and orchestration.undo_on_error:
            kwargs['undo_success'] = cc.undo()
        kwargs['end_time'] = get_now()
        rse.update_orch_execution(**kwargs)
    except Exception as e:
        current_app.logger.exception("Exception while executing invocation command")
        kwargs.update(success=False, message=str(e))
        rse.update_orch_execution(**kwargs)
        try:
            db.session.rollback()
        except:
            pass

    finally:
        if scope_enabled:
            lock.unlock(Scope.ORCHESTRATION, applicant=applicant, servers=servers)

    return execution.id



import argparse
import json
import os
import shlex
from pprint import pprint

from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.python import PythonLexer
from pygments.lexers.shell import BashLexer
from pygments.lexers.special import TextLexer

import dimensigon.dshell.network as ntwrk
from dimensigon.domain.entities import ActionType
from dimensigon.dshell import validators as v, converters as c
from dimensigon.dshell.argparse_raise import ArgumentParserRaise
from dimensigon.dshell.helpers import get_history, exit_dshell
from dimensigon.dshell.output import dprint
from dimensigon.dshell.prompts.utils import prompt_parameter, code_extension


def code_lexer(action_type):
    if action_type == 'SHELL':
        lexer = PygmentsLexer(BashLexer)
    elif action_type == 'PYTHON':
        lexer = PygmentsLexer(PythonLexer)
    else:
        lexer = PygmentsLexer(TextLexer)
    return lexer


form = {
    "name": dict(history=InMemoryHistory()),
    "version": dict(history=InMemoryHistory()),
    "action_type": dict(validator=v.Choice([at.name for at in ActionType if at.name != 'NATIVE']),
                        history=InMemoryHistory()),
    "code": dict(edit=code_extension, history=InMemoryHistory(), converter=c.MultiLine),
    "expected_stdout": dict(multiline=True, mouse_support=True, history=InMemoryHistory(), converter=c.MultiLine),
    "expected_stderr": dict(multiline=True, mouse_support=True, history=InMemoryHistory(), converter=c.MultiLine),
    "expected_rc": dict(validator=v.Int, converter=c.Int, history=InMemoryHistory()),
    "schema": dict(edit='.yaml', validator=v.JSON, history=InMemoryHistory(), converter=c.Yaml),
    "system_kwargs": dict(edit='.yaml', validator=v.JSON, history=InMemoryHistory(), converter=c.Yaml),
    "pre_process": dict(edit='.py', history=InMemoryHistory(), converter=c.MultiLine),
    "post_process": dict(edit='.py', history=InMemoryHistory(), converter=c.MultiLine),
}

parser = ArgumentParserRaise(allow_abbrev=False, prog='')
subparser = parser.add_subparsers(dest='cmd')
preview_action = subparser.add_parser('preview')
preview_action.set_defaults(func=lambda x: pprint(x))
set_action = subparser.add_parser('set')
set_action.add_argument('parameter', choices=form.keys())
delete_parser = subparser.add_parser('delete')
delete_parser.add_argument('parameter', choices=form.keys())
submit_action = subparser.add_parser('submit')
dump_action = subparser.add_parser('dump')
dump_action.add_argument('file')
exit_action = subparser.add_parser('exit')

history = None

entity_name = os.path.basename(__file__).rstrip('.py')


def subprompt(entity, changed=False, ask_all=False, parent_prompt=None):
    global history
    if history is None:
        history = get_history(entity_name, InMemoryHistory())

    if ask_all:
        for k, v in form.items():
            if k not in entity:
                try:
                    if prompt_parameter(k, entity, entity_name, form, parent_prompt):
                        changed = True
                except KeyboardInterrupt:
                    entity[k] = None
                except EOFError:
                    exit_dshell()

    while True:
        try:
            text = prompt(f"{parent_prompt}{entity_name}('{entity['name']}')> ", history=history,
                          auto_suggest=AutoSuggestFromHistory())
        except KeyboardInterrupt:
            continue
        except EOFError:
            exit_dshell(rc=1)
        try:
            namespace = parser.parse_args(shlex.split(text))
        except (ValueError, argparse.ArgumentError) as e:
            print(e)
            continue
        except SystemExit:
            continue

        if namespace.cmd == 'preview':
            dprint(entity)
        elif namespace.cmd == 'set':
            if namespace.parameter not in form.keys():
                dprint("Not a valid parameter. Available: " + ', '.join(form.keys()))
            else:
                try:
                    if prompt_parameter(namespace.parameter, entity, form,
                                        f"{parent_prompt}{entity_name}('{entity['name']}')"):
                        changed = True
                except KeyboardInterrupt:
                    continue  # Control-C pressed. Try again.
                except EOFError:
                    exit_dshell(rc=1)
        elif namespace.cmd == 'submit':
            resp = ntwrk.post(f'api_1_0.{entity_name.replace("_", "")}list', json=entity)
            dprint(resp)
            if resp.ok:
                return
        elif namespace.cmd == 'delete':
            if namespace.parameter not in form.keys():
                dprint("Not a valid parameter. Available: " + ', '.join(form.keys()))
            else:
                entity.pop(namespace.parameter)
                changed = True
        elif namespace.cmd == 'dump':
            with open(namespace.file, 'w') as dumpfile:
                json.dump(entity, dumpfile, indent=4)
            changed = False
        elif namespace.cmd == 'exit':
            if changed:
                text = ''
                while text.lower().strip() not in ('y', 'n'):
                    text = prompt('If you exit you will lose the changes. Do you want to continue? (y/n): ')
                if text.lower().strip() == 'y':
                    break
            else:
                break



import abc
import inspect
import random
import string
import uuid
from datetime import datetime

from sqlalchemy import Column, Boolean

from dimensigon import defaults
from dimensigon.utils.typos import UUID, UtcDateTime
from dimensigon.web.helpers import QueryWithSoftDelete


class SoftDeleteMixin:
    __prefix__ = '_old_'
    deleted = Column(Boolean(), default=False)
    query_class = QueryWithSoftDelete

    def __init__(self, deleted=False, **kwargs):
        # def wrapper_to_json(_self, func):
        #     @functools.wraps(func)
        #     def wrapper(*args, **kwargs):
        #         no_delete = kwargs.pop('no_delete', False)
        #         dto = func(*args, **kwargs)
        #         if not no_delete:
        #             dto.update({'deleted': _self.deleted})
        #             for attr in [attr for attr, value in inspect.getmembers(_self) if
        #                          attr.startswith(_self.__prefix__)]:
        #                 dto.update({attr: getattr(_self, attr)})
        #         return dto
        #
        #     sig = inspect.signature(func)
        #     param_list = list(sig.parameters.values())
        #     try:
        #         v_p_i = max(
        #             [param_list.index(p) for p in param_list if p.kind == inspect.Parameter.VAR_POSITIONAL])
        #     except ValueError:
        #         v_p_i = None
        #     try:
        #         var_k_i = min([param_list.index(p) for p in param_list if p.kind == inspect.Parameter.VAR_KEYWORD])
        #     except ValueError:
        #         var_k_i = None
        #
        #     # determine position
        #     if var_k_i is not None:
        #         index = var_k_i - 1
        #     else:
        #         index = len(param_list)
        #     # determine type
        #     if v_p_i is None:
        #         kind = inspect.Parameter.POSITIONAL_OR_KEYWORD
        #     else:
        #         kind = inspect.Parameter.KEYWORD_ONLY
        #     if 'no_delete' not in sig.parameters:
        #         no_delete = inspect.Parameter("no_delete", kind=kind, default=False)
        #         param_list.insert(index, no_delete)
        #         new_sig = sig.replace(parameters=param_list)
        #         wrapper.__signature__ = new_sig
        #         return wrapper
        #     else:
        #         return func

        super().__init__(**kwargs)
        self.deleted = deleted
        for attr, value in kwargs.items():
            if attr.startswith(self.__prefix__):
                setattr(self, attr, kwargs.get(attr, None))
        # if hasattr(self, 'to_json') and callable(getattr(self, 'to_json')):
        #     self.to_json = wrapper_to_json(self, self.to_json)

    def to_json(self, no_delete=False, **kwargs):
        if hasattr(super(), 'to_json'):
            dto = super().to_json(**kwargs)
        else:
            dto = {}
        if not no_delete:
            dto.update({'deleted': self.deleted})
            for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
                dto.update({attr: getattr(self, attr)})
        return dto

    def delete(self):
        if not self.deleted:
            self.deleted = True
            for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
                original_attr = attr.lstrip(self.__prefix__)
                setattr(self, attr, getattr(self, original_attr))
                setattr(self, original_attr,
                        ''.join(random.choices(string.digits + string.ascii_letters + string.punctuation, k=10)))


class DistributedEntityMixin:
    order = None
    last_modified_at = Column(UtcDateTime(), nullable=False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.last_modified_at = kwargs.pop('last_modified_at', None)

    def to_json(self, **kwargs):
        if hasattr(super(), 'to_json'):
            dto = super().to_json(**kwargs)
        else:
            dto = {}
        if self.last_modified_at:
            dto.update(last_modified_at=self.last_modified_at.strftime(defaults.DATEMARK_FORMAT))
        return dto

    @classmethod
    def from_json(cls, kwargs):
        if 'last_modified_at' in kwargs:
            last_modified_at = kwargs.pop('last_modified_at')
            last_modified_at = datetime.strptime(last_modified_at, defaults.DATEMARK_FORMAT)
            kwargs.update(last_modified_at=last_modified_at)


class UUIDEntityMixin:
    id = Column(UUID, primary_key=True, default=lambda: str(uuid.uuid4()))

    def __init__(self, *args, id=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = id or str(uuid.uuid4())

    def to_json(self, **kwargs):
        if hasattr(super(), 'to_json'):
            dto = super().to_json(**kwargs)
        else:
            dto = {}
        if self.id:
            dto.update(id=str(self.id))
        return dto


class UUIDistributedEntityMixin(UUIDEntityMixin, DistributedEntityMixin):

    @classmethod
    @abc.abstractmethod
    def from_json(cls, kwargs):
        super().from_json(kwargs)
        try:
            o = cls.query.get(kwargs.get('id'))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)


class EntityReprMixin:
    id = None

    def __repr__(self):
        if self.id:
            return f'<{self.__class__.__name__} {self.id}>'
        else:
            return f'<{self.__class__.__name__} (transient {id(self)})>'

    def __str__(self):
        return self.__repr__()



# business functions related with business logic
import base64
import logging
import math
import os
import subprocess
import typing as t

import aiohttp
from flask import current_app

from dimensigon import defaults
from dimensigon.domain.entities import Server
from dimensigon.utils import asyncio
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, errors
from dimensigon.web.network import get

catalog_logger = logging.getLogger('dm.catalog')


def run_elevator(file, new_version, logger):
    logger.info(f"Upgrading to version {new_version}")
    stdout = open('elevator.out', 'a')
    cmd = ['python', 'elevator.py', 'upgrade', file, str(new_version)]
    logger.debug(f"Running command {' '.join(cmd)}")
    subprocess.Popen(cmd, stdin=None, stdout=stdout, stderr=stdout, close_fds=True, env=os.environ)
    stdout.close()


def get_software(server: Server, auth) -> t.Tuple[str, str]:
    resp, code = get(server, 'api_1_0.software_dimensigon', auth=auth)
    if code == 200:
        content = base64.b64decode(resp.get('content').encode('ascii'))

        file = os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon', resp.get('filename'))
        with open(file, 'wb') as fh:
            fh.write(content)
        return file, resp.get('version')
    else:
        return None, None


async def async_send_file(dest_server: Server, transfer_id: Id, file,
                          chunk_size: int = None, chunks: int = None, max_senders: int = None,
                          identity=None, retries: int = 3):
    async def send_chunk(server: Server, view: str, _chunk, _chunk_size, sem, _session):
        async with sem:
            json_msg = dict(chunk=_chunk)

            with open(file, 'rb') as fd:
                fd.seek(_chunk * _chunk_size)
                chunk_content = base64.b64encode(fd.read(_chunk_size)).decode('ascii')
            json_msg.update(content=chunk_content)

            return await ntwrk.async_post(server, view_or_url=view,
                                          view_data=dict(transfer_id=str(transfer_id)), json=json_msg,
                                          session=_session, identity=identity)

    chunk_size = chunk_size or defaults.CHUNK_SIZE
    max_senders = max_senders or defaults.MAX_SENDERS
    chunks = chunks or math.ceil(os.path.getsize(file) / chunk_size)
    retries = retries
    sem = asyncio.Semaphore(max_senders)
    l_chunks = [c for c in range(0, chunks)]
    async with aiohttp.ClientSession() as session:
        while retries > 0:
            responses = {}
            retry_chunks = []
            for chunk in l_chunks:
                task = asyncio.create_task(
                    send_chunk(dest_server, 'api_1_0.transferresource', chunk, chunk_size, sem, session))
                responses.update({chunk: task})

            for chunk, task in responses.items():
                resp = await task
                if resp.code != 201:
                    retry_chunks.append(chunk)
                elif resp.code == 410:
                    raise errors.TransferNotInValidState(transfer_id, resp.msg['error'].get(['status'], None))

            if len(retry_chunks) == 0 and chunks != 1:
                resp = await ntwrk.async_put(dest_server, 'api_1_0.transferresource',
                                             view_data={'transfer_id': transfer_id},
                                             session=session,
                                             identity=identity)
                if resp.code != 201:
                    current_app.logger.error(
                        f"Transfer {transfer_id}: Unable to create file at destination {dest_server.name}: "
                        f"{resp}")
            l_chunks = retry_chunks
            retries -= 1
    if l_chunks:
        data = {c: responses[c].result() for c in l_chunks}
        resp = await ntwrk.async_patch(dest_server, 'api_1_0.transferresource',
                                       view_data={'transfer_id': transfer_id},
                                       json={'status': 'TRANSFER_ERROR'},
                                       session=session,
                                       identity=identity)
        if not resp.ok:
            current_app.logger.error(f"Transfer {transfer_id}: Unable to change state to TRANSFER_ERROR")
        raise errors.ChunkSendError(data)


import base64
import logging
import os
import queue
import time
import typing as t
import zlib
from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor

from dataclasses import dataclass
from sqlalchemy import orm
from sqlalchemy.orm import sessionmaker
from watchdog.events import FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import ObservedWatch

from dimensigon import defaults
from dimensigon.domain.entities import File, Server, Log, FileServerAssociation
from dimensigon.domain.entities.log import Mode
from dimensigon.use_cases.cluster import NewEvent, AliveEvent
from dimensigon.use_cases.mptools import MPQueue, TimerWorker
from dimensigon.utils import asyncio
from dimensigon.utils.helpers import remove_root
from dimensigon.utils.pygtail import Pygtail
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, get_root_auth

if t.TYPE_CHECKING:
    from dimensigon.core import Dimensigon

_logger = logging.getLogger('dm.FileSync')
_log_logger = logging.getLogger('dm.logfed')

MAX_LINES = 10000  # max lines readed from a log
# period of time process checks for new files added to the database. must be equal or bigger than defaults.
FILE_WATCHES_REFRESH_PERIOD = 30
MAX_ALLOWED_ERRORS = 2  # max allowed errors to consider a node blacklisted
RETRY_BLACKLIST = 300  # retry blacklisted servers after RETRY_BLACKLIST seconds


class _PygtailBuffer(Pygtail):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._buffer = None

    def fetch(self):
        if self._buffer:
            return self._buffer
        else:
            self._buffer = ''.join(self.readlines(max_lines=MAX_LINES))
        return self._buffer

    def update_offset_file(self):
        self._buffer = None
        super().update_offset_file()


FileWatchId = t.Tuple[Id, str]


class WatchdogEventHandler(PatternMatchingEventHandler):

    def __init__(self, fw_id: FileWatchId, fs: 'FileSync', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.fw_id = fw_id
        self.fs = fs
        self.last_modified = time.time()

    def on_any_event(self, event: FileSystemEvent):
        if time.time() - self.last_modified > 1:
            _logger.debug(f"{event.event_type} event triggered on {event.src_path}")
            self.last_modified = time.time()
            self.fs.add(self.fw_id)


@dataclass
class BlacklistEntry:
    retries: int = 0
    blacklisted: float = None


class FileSync(TimerWorker):
    ###########################
    # START Class Inheritance #
    def init_args(self, dimensigon: 'Dimensigon', file_sync_period=defaults.FILE_SYNC_PERIOD,
                  file_watches_refresh_period=FILE_WATCHES_REFRESH_PERIOD, max_allowed_errors=MAX_ALLOWED_ERRORS,
                  retry_blacklist=RETRY_BLACKLIST):
        self.dm = dimensigon

        # Multiprocessing
        self.queue = MPQueue()

        # Parameters
        self.INTERVAL_SECS = file_sync_period
        self.file_watches_refresh_period = file_watches_refresh_period
        self.max_allowed_errors = max_allowed_errors
        self.retry_blacklist = retry_blacklist

        # internals
        self._changed_files: t.Set[Id] = set()  # list of changed files to be sent
        self._changed_servers: t.Dict[Id, t.List[Id]] = dict()
        self._file2watch: t.Dict[Id, ObservedWatch] = {}
        self._last_file_updated = None
        self._blacklist: t.Dict[t.Tuple[Id, Id], BlacklistEntry] = {}
        self._blacklist_log: t.Dict[t.Tuple[Id, Id], BlacklistEntry] = {}
        self.session = None
        self._server = None
        self._loop = None

        # log variables
        self._mapper: t.Dict[Id, t.List[_PygtailBuffer]] = {}

    def startup(self):
        self._executor = ThreadPoolExecutor(max_workers=max(os.cpu_count(), 4),
                                            thread_name_prefix="FileSyncThreadPool")
        self._observer = Observer()
        self.session = self._create_session()
        self._observer.start()

        self._set_initial_modifications()
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self.dispatcher.listen([NewEvent, AliveEvent], lambda x: self.add(None, x.args[0]))

    def shutdown(self):
        self.session.close()
        self._observer.stop()
        self._executor.shutdown()

    def main_func(self):
        # collect new File events
        while True:
            item = self.queue.safe_get()
            if item:
                self._add(*item)
            else:
                break
        self._set_watchers()
        self._sync_files()

        # send log data
        self._send_new_data()

    # END Class Inheritance #
    #########################

    ############################
    # INIT Interface functions #
    def add(self, file: t.Union[File, Id] = None, server: t.Union[Server, Id] = None):
        if isinstance(file, File):
            file_id = file.id
        else:
            file_id = file
        if isinstance(server, Server):
            server_id = server.id
        else:
            server_id = server
        try:
            self.queue.safe_put((file_id, server_id), timeout=2)
        except queue.Full:
            self.logger.warning("Queue is full. Try increasing its size")

    # END Interface functions  #
    ############################

    ##############################
    # INNER methods & attributes #
    def _create_session(self):
        self.Session = sessionmaker(bind=self.dm.engine)
        return self.Session()

    @property
    def server(self) -> Server:
        if self._server is None:
            self._server = self.session.query(Server).filter_by(_me=1, deleted=0).one_or_none()
        return self._server

    @property
    def my_files_query(self):
        return self.session.query(File).filter_by(deleted=0).filter_by(
            src_server_id=self.server.id) if self.session else None
        # return File.query.filter_by(src_server_id=self._server_id)

    @property
    def file_query(self):
        return self.session.query(File).filter_by(deleted=0) if self.session else None
        # return File.query

    def get_file(self, file_id):
        return self.file_query.filter_by(id=file_id).one_or_none() if self.session else None
        # return self.file_query.filter_by(id=file_id).one_or_none()

    def _add(self, file_id: Id = None, server_id: Id = None):
        if server_id:
            if file_id is None:
                # new server alive, send all files
                for fsa in self.session.query(FileServerAssociation).filter_by(dst_server_id=server_id, deleted=0).join(
                        File).filter_by(src_server_id=self.server.id).all():
                    if not (getattr(fsa.file, 'deleted', True) or fsa.destination_server.deleted):
                        mtime = os.stat(fsa.target).st_mtime_ns
                        if fsa.l_mtime != mtime:
                            self._add(fsa.file.id, fsa.destination_server.id)
            elif file_id not in self._changed_files:
                # send data to a specific server
                if file_id not in self._changed_servers:
                    self._changed_servers[file_id] = [server_id]
                else:
                    if server_id not in self._changed_servers[file_id]:
                        self._changed_servers[file_id].append(server_id)
        else:
            self._changed_files.add(file_id)
            if file_id in self._changed_servers:
                self._changed_servers.pop(file_id)

    @staticmethod
    def _read_file(file, compress=True):
        with open(file, 'rb') as fd:
            if compress:
                return base64.b64encode(zlib.compress(fd.read())).decode('utf-8')
            else:
                return base64.b64encode(fd.read()).decode('utf-8')

    async def _send_file(self, file: File, servers: t.List[Id] = None):
        try:
            content = await self._loop.run_in_executor(self._executor, self._read_file, file.target)
        except Exception as e:
            self.logger.exception(f"Unable to get content from file {file.target}.")
            return

        if servers:
            server_ids = servers
            fsas = [fsa for fsa in file.destinations if fsa.destination_server.id in server_ids]
        else:
            server_ids = [fsa.destination_server.id for fsa in file.destinations]
            fsas = file.destinations

        with self.dm.flask_app.app_context():
            auth = get_root_auth()
            alive = self.dm.cluster_manager.get_alive()
            tasks = [ntwrk.async_post(fsa.destination_server, view_or_url='api_1_0.file_sync',
                                      view_data={'file_id': file.id},
                                      json=dict(file=fsa.target, data=content, force=True),
                                      auth=auth) for fsa in fsas if fsa.destination_server.id in alive]
            skipped = [fsa.destination_server.name for fsa in fsas if fsa.destination_server.id not in alive]
            if skipped:
                self.logger.debug(
                    f"Following servers are skipped because we do not see them alive: {', '.join(skipped)}")
            if tasks:
                self.logger.debug(
                    f"Syncing file {file} with the following servers: {', '.join([fsa.destination_server.name for fsa in fsas if fsa.destination_server.id in alive])}.")

                resp = await asyncio.gather(*tasks)
                for resp, fsa in zip(resp, fsas):
                    if not resp.ok:
                        self.logger.warning(
                            f"Unable to send file {file.target} to {fsa.destination_server}. Reason: {resp}")
                        if (file.id, fsa.destination_server.id) not in self._blacklist:
                            bl = BlacklistEntry()
                            self._blacklist[(file.id, fsa.destination_server.id)] = bl
                        else:
                            bl = self._blacklist.get((file.id, fsa.destination_server.id))
                        bl.retries += 1
                        if bl.retries >= self.max_allowed_errors:
                            self.logger.debug(f"Adding server {fsa.destination_server} to the blacklist.")
                            bl.blacklisted = time.time()
                    else:
                        if (file.id, fsa.destination_server.id) in self._blacklist:
                            self._blacklist.pop((file.id, fsa.destination_server.id), None)
                        fsa.l_mtime = file.l_mtime
                try:
                    self.session.commit()
                except:
                    self.session.rollback()

    def _sync_files(self):
        coros = []
        for file_id in self._changed_files:
            f = self.get_file(file_id)
            if f:
                try:
                    f.l_mtime = os.stat(f.target).st_mtime_ns
                except FileNotFoundError:
                    pass
                else:
                    coros.append(self._send_file(f))
        if coros:
            try:
                self.session.commit()
            except:
                self.session.rollback()

        for file_id, server_id in list(self._blacklist.keys()):
            file = self.get_file(file_id)
            dest = self.session.query(Server).filter_by(deleted=0).filter_by(id=server_id).count()
            bl = self._blacklist[(file_id, server_id)]
            if file and dest:  # file and server in the black list may be deleted
                # if server_id in getattr(getattr(self.app, 'cluster_manager', None), 'cluster', [server_id]):
                if bl.retries < self.max_allowed_errors or time.time() - bl.blacklisted > self.retry_blacklist:
                    if file_id not in self._changed_files:
                        coros.append(self._send_file(file, [server_id]))
            else:
                self._blacklist.pop((file_id, server_id), None)

        for file_id, server_ids in self._changed_servers.items():
            file = self.get_file(file_id)
            if file:
                coros.append(self._send_file(file, server_ids))

        if coros:
            try:
                self._loop.run_until_complete(asyncio.gather(*coros, return_exceptions=False))
            except Exception:
                self.logger.exception("Error while trying to send data.")
        self._changed_files.clear()
        self._changed_servers.clear()

    def _schedule_file(self, file_id, target=None):
        if isinstance(file_id, File):
            target = file_id.target
            file_id = file_id.id
        assert target is not None
        weh = WatchdogEventHandler(file_id, self, patterns=[target])
        try:
            watch = self._observer.schedule(weh, os.path.dirname(target), recursive=False)
        except FileNotFoundError:
            pass
        else:
            self._file2watch.update({file_id: watch})

    def _set_initial_modifications(self):
        for file in self.my_files_query.all():
            self._schedule_file(file)
            try:
                if os.path.exists(file.target):
                    mtime = os.stat(file.target).st_mtime_ns
                    if mtime != file.l_mtime:
                        self._add(file.id, None)
                    else:
                        for fsa in file.destinations:
                            if fsa.l_mtime != mtime:
                                self._add(file.id, fsa.destination_server.id)
            except:
                pass

    def _set_watchers(self, force=False):
        if self._last_file_updated is None or time.time() - self._last_file_updated > self.file_watches_refresh_period or force:
            self._last_file_updated = time.time()
            id2target = {f.id: f.target for f in self.my_files_query.options(orm.load_only("id", "target")).all()}
            files_from_db = set(id2target.keys())
            files_already_watching = set(self._file2watch.keys())

            file_watches_to_remove = files_already_watching - files_from_db
            file_watches_to_add = files_from_db - files_already_watching

            if file_watches_to_remove:
                self.logger.debug(f"Unscheduling the following files: {file_watches_to_remove}")
                for file_id in file_watches_to_remove:
                    self._observer.unschedule(self._file2watch[file_id])
                    self._file2watch.pop(file_id)

            if file_watches_to_add:
                self.logger.debug(f"Scheduling the following files: {file_watches_to_add}")
                for file_id in file_watches_to_add:
                    self._schedule_file(file_id, id2target[file_id])
                    # add for sending file for first time
                    self._add(file_id, None)

    @property
    def my_logs(self):
        return self.session.query(Log).filter_by(source_server=Server.get_current(session=self.session)).all()

    def update_mapper(self):
        logs = self.my_logs
        id2log = {log.id: log for log in logs}
        # remove logs
        for log_id in list(self._mapper.keys()):
            if log_id not in id2log:
                del self._mapper[log_id]

        # add new logs
        for log in logs:
            if log.id not in self._mapper:
                self._mapper[log.id] = []
            self.update_pytail_objects(log, self._mapper[log.id])

    def update_pytail_objects(self, log: Log, pytail_list: t.List):
        if os.path.isfile(log.target):
            if len(pytail_list) == 0:
                filename = '.' + os.path.basename(log.target) + '.offset'
                path = os.path.dirname(log.target)
                offset_file = self.dm.config.path(defaults.OFFSET_DIR, remove_root(path), filename)
                if not os.path.exists(offset_file):
                    _log_logger.debug(f"creating offset file {offset_file}")
                    os.makedirs(os.path.dirname(offset_file), exist_ok=True)
                pytail_list.append(
                    _PygtailBuffer(file=log.target, offset_mode='manual', offset_file=offset_file))
        else:
            for folder, dirnames, filenames in os.walk(log.target):
                for filename in filenames:
                    if log._re_include.search(filename) and not log._re_exclude.search(filename):
                        file = os.path.join(folder, filename)
                        offset_file = os.path.join(folder, '.' + filename + '.offset')
                        if not any(map(lambda p: p.file == file, pytail_list)):
                            pytail_list.append(
                                _PygtailBuffer(file=file, offset_mode='manual', offset_file=offset_file))
                if not log.recursive:
                    break
                new_dirnames = []
                for dirname in dirnames:
                    if log._re_include.search(dirname) and not log._re_exclude.search(dirname):
                        new_dirnames.append(dirname)
                dirnames[:] = new_dirnames

    def _send_new_data(self):
        self.update_mapper()
        tasks = OrderedDict()

        for log_id, pb in self._mapper.items():
            log = self.session.query(Log).get(log_id)
            for pytail in pb:
                data = pytail.fetch()
                data = data.encode() if isinstance(data, str) else data
                if data and log.destination_server.id in self.dm.cluster_manager.get_alive():
                    if log.mode == Mode.MIRROR:
                        file = pytail.file
                    elif log.mode == Mode.REPO_ROOT:
                        path_to_remove = os.path.dirname(log.target)
                        relative = os.path.relpath(pytail.file, path_to_remove)
                        file = os.path.join('{LOG_REPO}', relative)
                    elif log.mode == Mode.FOLDER:
                        path_to_remove = os.path.dirname(log.target)
                        relative = os.path.relpath(pytail.file, path_to_remove)
                        file = os.path.join(log.dest_folder, relative)
                    else:
                        def get_root(dirname):
                            new_dirname = os.path.dirname(dirname)
                            if new_dirname == dirname:
                                return dirname
                            else:
                                return get_root(new_dirname)

                        relative = os.path.relpath(pytail.file, get_root(pytail.file))
                        file = os.path.join('{LOG_REPO}', relative)
                    with self.dm.flask_app.app_context():
                        auth = get_root_auth()

                    task = ntwrk.async_post(log.destination_server, 'api_1_0.logresource',
                                            view_data={'log_id': str(log_id)},
                                            json={"file": file,
                                                  'data': base64.b64encode(zlib.compress(data)).decode('ascii'),
                                                  "compress": True},
                                            auth=auth)

                    tasks[task] = (pytail, log)
                    _log_logger.debug(f"Task sending data from '{pytail.file}' to '{log.destination_server}' prepared")

        if tasks:
            with self.dm.flask_app.app_context():
                responses = asyncio.run(asyncio.gather(*list(tasks.keys())))

            for task, resp in zip(tasks.keys(), responses):
                pytail, log = tasks[task]
                if resp.ok:
                    pytail.update_offset_file()
                    _log_logger.debug(f"Updated offset from '{pytail.file}'")
                    if log.id not in self._blacklist:
                        self._blacklist_log.pop(log.id, None)
                else:
                    _log_logger.error(
                        f"Unable to send log information from '{pytail.file}' to '{log.destination_server}'. Error: {resp}")
                    if log.id not in self._blacklist:
                        bl = BlacklistEntry()
                        self._blacklist_log[log.id] = bl
                    else:
                        bl = self._blacklist_log.get(log.id)
                    bl.retries += 1
                    if bl.retries >= self.max_allowed_errors:
                        _log_logger.debug(f"Adding server {log.destination_server.id} to the blacklist.")
                        bl.blacklisted = time.time()



import time


class EventMessage:
    __slots__ = ('__id', '__event_type', '__source', 'args', 'kwargs')

    def __init__(self, __event_type: str, *args, source: str = None, **kwargs):
        self.__id = time.time()
        self.__event_type = __event_type
        self.__source = source
        self.args = tuple(args)
        self.kwargs = kwargs

    @property
    def id(self):
        return self.__id

    @property
    def event_type(self):
        return self.__event_type

    @property
    def source(self):
        return self.__source

    def __str__(self):
        s = ""
        if self.__source:
            s += f"{self.__source}-"
        s += f"{self.__event_type}("
        if self.args:
            s += str(self.args)[1:-1]
        if self.kwargs:
            s += ', ' + str(self.kwargs)[1:-1]
        s += ')'
        return s

    def __repr__(self):
        return str(self)


class BaseEvent(EventMessage):

    def __init__(self, *args, source: str = None, **kwargs):
        super().__init__(self.__class__.__name__, *args, source=source, **kwargs)


class Stop(BaseEvent):
    """Event to stop the program"""


class StopEventHandler(BaseEvent):
    """Event to stop """



import json
import typing as t

import yaml


class Converter:

    @staticmethod
    def load(text) -> t.Any:
        return text

    @staticmethod
    def dump(data) -> str:
        return str(data)


class JSON(Converter):

    @staticmethod
    def load(text):
        return json.loads(text)

    @staticmethod
    def dump(data):
        return json.dumps(data)


class Bool(Converter):
    true = ['true', 'yes', 'y']
    false = ['false', 'no', 'n']

    @classmethod
    def load(cls, x: str):
        if x.strip().lower() in cls.true:
            return True
        elif x.strip().lower() in cls.false:
            return False
        else:
            raise ValueError

    @classmethod
    def dump(cls, data):
        if data:
            return cls.true[0]
        else:
            return cls.false[0]


class Int(Converter):

    @staticmethod
    def load(x: str):
        return int(x) if x is not None else x

    @staticmethod
    def dump(data):
        return str(data)


class List(Converter):

    def __init__(self, separator=None) -> None:
        self.separator = separator

    @staticmethod
    def dump(data):
        return ' '.join(data or [])

    def load(self, x: str):
        return x.split(sep=self.separator)


class MultiLine(Converter):

    @staticmethod
    def load(text) -> t.Any:
        return text.split('\n')

    @staticmethod
    def dump(data: t.List) -> str:
        return '\n'.join(data)


class Yaml(Converter):

    @staticmethod
    def load(text):
        return yaml.load(text, Loader=yaml.FullLoader)

    @staticmethod
    def dump(data: dict) -> str:
        if data:
            return yaml.dump(data, Dumper=yaml.Dumper)
        else:
            return ""



import datetime as dt
import json
import logging
import random
import threading
import typing as t

from dataclasses import dataclass
from sqlalchemy import orm
from sqlalchemy.orm import sessionmaker

from dimensigon import defaults
from dimensigon.domain.entities import Server, Route
from dimensigon.use_cases.mptools import Worker, MPQueue
from dimensigon.use_cases.mptools_events import BaseEvent
from dimensigon.use_cases.routing import InitialRouteSet
from dimensigon.utils import asyncio
from dimensigon.utils.helpers import format_exception, get_now
from dimensigon.utils.typos import Id
from dimensigon.web import network as ntwrk, get_root_auth

if t.TYPE_CHECKING:
    from dimensigon.core import Dimensigon


# self.logger = logging.getLogger('dm.cluster')


# async def send_cluster_register(cr, servers=None, auth=None, exclude=None):
#     if not servers:
#         servers = Server.get_neighbours(exclude=exclude)
#     responses = await ntwrk.parallel_requests(servers, 'post',
#                                               view_or_url='api_1_0.cluster',
#                                               json=cr,
#                                               auth=auth or get_root_auth(),
#                                               timeout=10)
# 
#     for r, s in zip(responses, servers):
#         if not r.ok and self.logger.level <= logging.WARNING:
#             self.logger.warning(
#                 f"Unable to send cluster information to {s}. Response: {r}")
# 
# 
# def check_server_alive(server: Server):
#     alive_server_ids = [i for i in current_app.dm.cluster_manager.get_alive() if
#                         server.id != i and i != Server.get_current().id]
#     # check if I have it as a neighbour
#     if server.route and server.route.cost == 0:
#         route = routing.check_gates(server)
#         if route:
#             return True
# 
#     # in order to prevent broadcast to everyone, first try a ping
#     cost, elapsed = ntwrk.ping(server, retries=1, timeout=15)
#     if cost is not None:
#         return True
#     responses = asyncio.run(ntwrk.parallel_requests(alive_server_ids, 'get',
#                                                     view_or_url='api_1_0.routes_neighbour',
#                                                     view_data=dict(server_id=server.id), timeout=10))
#     for r in responses:
#         if r.ok:
#             if r.msg.get('neighbour'):
#                 return True
#     return False


# def update_cluster_status():
#     alive_server_ids = current_app.dm.cluster_manager.get_alive()
#     updated = False
#     for alive_server_id in alive_server_ids:
#         if alive_server_id != Server.get_current().id:
#             alive = check_server_alive(Server.query.get(Server))
#             if not alive:
#                 current_app.dm.cluster_manager.put(ident=alive_server_id, keepalive=get_now(), death=True)
#                 updated = True
#     return updated
# 
# 
# async def check_heartbeat_and_send(cluster_session_id: int, heartbeat_id: dt.datetime,
#                                    exclude: t.Optional[t.List[Id]] = None):
#     cr = current_app.dm.cluster_manager.put(ident=cluster_session_id, keepalive=heartbeat_id)
#     if cr:
#         await send_cluster_register(cr, exclude=exclude)


class ClusterEvent(BaseEvent):
    """An Base class for Cluster related Event"""


class DeathEvent(ClusterEvent):
    """A node changed its state to DEATH"""


class NewEvent(ClusterEvent):
    """A node changed its state to NEW"""


class KeepAliveEvent(ClusterEvent):
    """A new KeepAlive Event has arrived"""


class ZombieEvent(ClusterEvent):
    """A node changed its state to ZOMBIE"""


class AliveEvent(ClusterEvent):
    """A node that was in ZOMBIE state changed its state to ALIVE"""


class NotifyClusterEnded(ClusterEvent):
    """Notify Cluster Process Ended"""


@dataclass
class _Entry:
    id: Id
    keepalive: dt.datetime = None
    death: bool = False
    zombie: bool = False

Input = t.Tuple[Id, dt.datetime, bool, bool]
Item = t.Union[Input, t.List[Input]]


class ClusterManager(Worker):
    ###########################
    # START Class Inheritance #
    def init_args(self, dimensigon: 'Dimensigon', maxsize=None, zombie_threshold=defaults.ZOMBIE_NODE,
                  send_interval=defaults.CLUSTER_SEND_PERIOD):
        self.dm = dimensigon
        self.Session = sessionmaker(bind=self.dm.engine)
        self.queue = MPQueue(maxsize=maxsize or 10000)
        self._registry: t.Dict[Id, _Entry] = self.dm.manager.dict()
        # self._registry: t.Dict[Id, _Entry] = dict()

        self._timer_registry: t.Dict[Id, threading.Timer] = dict()
        self.zombie_threshold = dt.timedelta(seconds=zombie_threshold)

        self._buffer: t.Dict[Id, _Entry] = {}  # data to be sent
        self.send_interval = send_interval  # delay in seconds between data change and sending data to other nodes
        self._lock = threading.Lock()  # lock used for consistency with Timer threads
        self._change_buffer_lock = threading.RLock()
        self._timer = None

    def startup(self):
        self._route_initiated = threading.Event()
        self.dispatcher.listen(InitialRouteSet, lambda x: self._route_initiated.set())
        self.dispatcher.listen('Listening', lambda x: self._notify_cluster_in())
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

    def shutdown(self):
        self._notify_cluster_out()
        self.queue.close()
        self.queue.join_thread()
        for t in self._timer_registry.values():
            t.cancel()
        if self._timer:
            self._timer.cancel()

    def main_func(self, *args, **kwargs):
        item = self.queue.safe_get()
        if item:
            self._process_item(item)

    # END Class Inheritance #
    #########################

    ############################
    # INIT Interface functions #
    def put_many(self, data: t.List, block=True, timeout=None):
        self.queue.put(data, block, timeout)

    def put(self, ident: Id, keepalive: dt.datetime, death: bool = False, block=True, timeout=None):
        self.queue.put((ident, keepalive, death), block=block, timeout=timeout)

    def get_alive(self) -> t.List[Id]:
        return [e.id for e in self._registry.values() if not e.zombie and not e.death] + [self.dm.server_id]

    def get_zombies(self) -> t.List[Id]:
        return [e.id for e in self._registry.values() if e.zombie and not e.death]

    def get_cluster(self, str_format=None):
        return [(e.id, e.keepalive if str_format is None else e.keepalive.strftime(str_format), e.death) for e in
                self._registry.values()] + [
            (self.dm.server_id, get_now() if str_format is None else get_now().strftime(str_format), False)]

    def __contains__(self, item):
        entity = self._registry.get(item, None)
        if item == self.dm.server_id or (entity and not (entity.zombie or entity.death)):
            return True
        else:
            return False

    # END Interface functions  #
    ############################

    ##############################
    # INNER methods & attributes #
    def _add2buffer(self, entry: _Entry):
        with self._change_buffer_lock:
            self._buffer.update({entry.id: entry})
            if self._timer is None:
                self._timer = threading.Timer(interval=self.send_interval, function=self._send_data)
                self._timer.start()

    def _cancel_timer(self, ident):
        if ident in self._timer_registry:
            self._timer_registry[ident].cancel()
            del self._timer_registry[ident]

    def _set_timer(self, ident, keepalive):
        self._cancel_timer(ident)
        th = self._timer_registry[ident] = threading.Timer(interval=self.zombie_threshold.seconds,
                                                           function=self.queue.put,
                                                           args=((ident, keepalive, None, True),),
                                                           kwargs=dict(block=True))
        th.start()

    def _process_one(self, item: Input):
        item = _Entry(*item)
        # discard item if it's me
        if item.id != self.dm.server_id:
            with self._lock:
                current = self._registry.get(item.id,
                                             _Entry(item.id, keepalive=dt.datetime(1, 1, 1, tzinfo=dt.timezone.utc),
                                                    death=True, zombie=False))
                event = None
                if item.zombie:
                    # zombie message
                    if current.keepalive == item.keepalive and not current.death:
                        current.zombie = True
                        event = ZombieEvent(item.id)
                elif item.death:
                    # death message
                    if not current.death or (current.death and item.keepalive > current.keepalive):
                        self._cancel_timer(current.id)
                        current.death = True
                        current.zombie = False
                        current.keepalive = item.keepalive

                        self._add2buffer(current)
                        event = DeathEvent(current.id)

                else:
                    if item.keepalive > current.keepalive:
                        # alive
                        current.keepalive = item.keepalive
                        # restart timer
                        self._set_timer(item.id, item.keepalive)
                        if current.death:
                            current.death = False
                            current.zombie = False
                            event = NewEvent(item.id)
                        elif current.zombie:
                            current.zombie = False
                            event = AliveEvent(item.id)
                        else:
                            pass
                            # event = KeepAliveEvent(item.id, item.keepalive)
                        self._add2buffer(current)
                self._registry[item.id] = current
                self.publish_q.safe_put(event) if event else None

    def _process_item(self, item: Item):
        if isinstance(item, list):
            [self._process_one(i) for i in item]
        else:
            self._process_one(item)

    def _send_data(self):
        session = self.Session()

        def log_data(data):
            debug_data = []
            for cr in data:
                server = dict(id=cr.id)
                name = getattr(session.query(Server).get(cr.id), 'name', cr.id)
                if name:
                    server.update(name=name)

                debug_data.append(
                    {'server': server, 'keepalive': cr.keepalive.strftime(defaults.DATEMARK_FORMAT), 'death': cr.death})
            return debug_data

        # time to send data
        with self.dm.flask_app.app_context():
            neighbours = Server.get_neighbours(session=session)
            if neighbours:
                with self._change_buffer_lock:
                    temp_buffer = dict(self._buffer)
                    self._buffer.clear()

                self.logger.debug(
                    f"Sending cluster information to the following nodes: {', '.join([s.name for s in neighbours])}"
                )
                self.logger.log(1, f"{json.dumps(log_data(temp_buffer.values()), indent=2)}")

                auth = get_root_auth()
                try:
                    responses = asyncio.run(
                        ntwrk.parallel_requests(neighbours, 'POST', view_or_url='api_1_0.cluster',
                                                json=[{'id': e.id,
                                                       'keepalive': e.keepalive.strftime(defaults.DATEMARK_FORMAT),
                                                       'death': e.death} for e in
                                                      temp_buffer.values()], auth=auth, securizer=False), )
                except Exception as e:
                    self.logger.error(f"Unable to send cluster information to neighbours: {format_exception(e)}")
                    # restore data with new data arrived
                    with self._change_buffer_lock:
                        temp_buffer.update(**self._buffer)
                        self._buffer.clear()
                        self._buffer.update(temp_buffer)
                else:
                    for r in responses:
                        if not r.ok:
                            self.logger.warning(f"Unable to send data to {r.server}: {r}")

                # check if new data arrived during timer execution
                with self._change_buffer_lock:
                    if self._buffer:
                        self._timer = threading.Timer(interval=1, function=self._send_data)
                        self._timer.start()
                    else:
                        self._timer = None
            else:
                self.logger.debug(f"No neighbour servers to send cluster information")
                with self._change_buffer_lock:
                    self._timer = None
        session.close()

    def _notify_cluster_in(self):
        from dimensigon.domain.entities import Server
        import dimensigon.web.network as ntwrk
        from dimensigon.domain.entities import Parameter

        try:
            signaled = self._route_initiated.wait(timeout=120)
        except Exception:
            return

        if not signaled:
            self.logger.warning("Route Event not fired.")

        self.logger.debug("Notify Cluster")
        with self.dm.flask_app.app_context():
            not_notify = set()
            me = Server.get_current()

            msg = [r.to_json() for r in Route.query.options(orm.lazyload(Route.destination), orm.lazyload(Route.gate),
                                                            orm.lazyload(Route.proxy_server)).all()]

            neighbours = Server.get_neighbours()

            if Parameter.get('join_server', None):
                join_server = Server.query.get(Parameter.get('join_server'))
            else:
                join_server = None

            now = get_now()
            msg = dict(keepalive=now.strftime(defaults.DATEMARK_FORMAT), routes=msg)
            if neighbours:
                random.shuffle(neighbours)
                first = [s for s in neighbours if s.id == Parameter.get('new_gates_server', None)]
                if first:
                    neighbours.pop(neighbours.index(first[0]))
                    neighbours = first + neighbours
                elif join_server in neighbours:
                    neighbours.pop(neighbours.index(join_server))
                    neighbours = [join_server] + neighbours
                for s in neighbours:
                    if s.id not in not_notify:
                        self.logger.debug(f"Sending 'Cluster IN' message to {s}")
                        resp = ntwrk.post(s, 'api_1_0.cluster_in', view_data=dict(server_id=str(me.id)),
                                          json=msg, timeout=10, auth=get_root_auth())
                        if resp.ok:
                            converted = []
                            for ident, str_keepalive, death in resp.msg['cluster']:
                                try:
                                    keepalive = dt.datetime.strptime(str_keepalive, defaults.DATEMARK_FORMAT)
                                except ValueError:
                                    continue
                                converted.append((ident, keepalive, death))
                            self.put_many(converted)
                            not_notify.update(resp.msg.get('neighbours', []))
                        else:
                            self.logger.debug(f"Unable to send 'Cluster IN' message to {s} . Response: {resp}")
                    else:
                        self.logger.debug(f"Skiping server {s} from sending 'Cluster IN' message")
                # alive = [(getattr(Server.query.get(s_id), 'name', None) or s_id) for s_id in
                #          self.get_alive()]
                # self.logger.info(f"Alive servers: {', '.join(alive)}")
            else:
                self.logger.debug("No neighbour to send 'Cluster IN'")
        self.logger.debug("Notify Cluster ended")

    def _notify_cluster_out(self):
        with self.dm.flask_app.app_context():
            servers = Server.get_neighbours()
            if servers:
                self.logger.debug(f"Sending shutdown to {', '.join([s.name for s in servers])}")
            else:
                self.logger.debug("No server to send shutdown information")
            if servers:
                responses = asyncio.run(
                    ntwrk.parallel_requests(servers, 'post',
                                            view_or_url='api_1_0.cluster_out',
                                            view_data=dict(server_id=str(Server.get_current().id)),
                                            json={'death': get_now().strftime(defaults.DATEMARK_FORMAT)},
                                            timeout=2, auth=get_root_auth()))
                if self.logger.level <= logging.DEBUG:
                    for r in responses:
                        if not r.ok:
                            self.logger.warning(f"Unable to send data to {r.server}: {r}")



import functools
import multiprocessing as mp
import os
import signal
import sys

from dataclasses import dataclass


@dataclass
class Token:
    """
    Class that saves relation between a sent message and its asynchronous return.
    """
    id: int
    source: str
    destination: str

    @property
    def uid(self):
        return self.source + '.' + self.destination + '.' + str(self.id)


# -- Signal Handling
class TerminateInterrupt(BaseException):
    pass


class SignalObject:
    MAX_TERMINATE_CALLED = 3

    def __init__(self, shutdown_event):
        self.terminate_called = 0
        self.shutdown_event = shutdown_event


def default_signal_handler(signal_object, exception_class, signal_num, current_stack_frame):
    signal_object.terminate_called += 1
    signal_object.shutdown_event.set()
    if signal_object.terminate_called >= signal_object.MAX_TERMINATE_CALLED:
        raise exception_class()


def init_signal(signal_num, signal_object, exception_class, handler):
    handler = functools.partial(handler, signal_object, exception_class)
    signal.signal(signal_num, handler)
    signal.siginterrupt(signal_num, False)


def init_signals(shutdown_event, int_handler, term_handler):
    signal_object = SignalObject(shutdown_event)
    init_signal(signal.SIGINT, signal_object, KeyboardInterrupt, int_handler)
    init_signal(signal.SIGTERM, signal_object, TerminateInterrupt, term_handler)
    return signal_object


class Process:
    MAX_TERMINATE_CALLED = 3
    int_handler = staticmethod(default_signal_handler)
    term_handler = staticmethod(default_signal_handler)
    _logger = None

    def __init__(self, shutdown_event, name=None):
        self._stop = shutdown_event
        self.name = name or self.__class__.__name__
        self.process = mp.Process(target=self._run, name=self.name)

    def _init_signals(self):
        signal_object = init_signals(self._stop, self.int_handler, self.term_handler)
        return signal_object

    def _run(self):
        self._logger.info(f'Starting {self.name} ({os.getpid()})')
        if isinstance(self.process, mp.Process):
            self._init_signals()
        try:
            self._main()
        except (TerminateInterrupt, KeyboardInterrupt):
            sys.exit(1)
        except Exception:
            self._logger.exception(f"Error while executing {self.name}")
        finally:
            self._shutdown()
        self._logger.info(f"{self.name} stopped")

    def _main(self):
        raise NotImplemented

    def _shutdown(self):
        raise NotImplemented

    def start(self):
        self.process.start()



import copy
import os
import sys
import typing as t

from dimensigon.domain.entities import Server
from dimensigon.domain.entities.base import UUIDistributedEntityMixin, SoftDeleteMixin, DistributedEntityMixin
from dimensigon.utils import typos
from dimensigon.utils.typos import Id
from dimensigon.web import db
from dimensigon.web.helpers import QueryWithSoftDelete

if t.TYPE_CHECKING:
    ...

if sys.version_info < (3, 8):
    Destination_Server = t.Dict[str, t.Union[Id, Server, str]]
else:
    class Destination_Server(t.TypedDict):
        dst_server_id: Id
        dest_folder: str

Destination_Servers = t.List[t.Union[t.Tuple[Server, str], t.Tuple[Id, str], Server, Destination_Server]]


class FileServerAssociation(DistributedEntityMixin, SoftDeleteMixin, db.Model):
    __tablename__ = 'D_file_server_association'
    order = 30

    file_id = db.Column(typos.UUID, db.ForeignKey('D_file.id'), nullable=False, primary_key=True)
    dst_server_id = db.Column(typos.UUID, db.ForeignKey('D_server.id'), nullable=False, primary_key=True)
    dest_folder = db.Column(db.Text)
    l_mtime = db.Column(db.INTEGER)

    file = db.relationship("File")
    destination_server = db.relationship("Server")

    @property
    def destination_folder(self):
        return self.dest_folder or self.file.dest_folder or os.path.dirname(self.file.target)

    @property
    def target(self):
        return os.path.join(self.destination_folder, os.path.basename(self.file.target))

    def to_json(self, human=False, **kwargs) -> t.Dict:
        data = super().to_json(**kwargs)
        if human:
            data.update({'file': {'target': self.file.target, 'src_server': self.file.source_server.name,
                                  'dst_server': str(self.destination_server.name),
                                  'dest_folder': self.dest_folder}})
        else:
            data.update({'file_id': str(self.file.id), 'dst_server_id': str(self.destination_server.id),
                         'dest_folder': self.dest_folder})
        return data

    @classmethod
    def from_json(cls, kwargs) -> 'FileServerAssociation':
        kwargs = copy.deepcopy(kwargs)
        kwargs['file'] = File.query.get_or_raise(kwargs.get('file_id'))
        kwargs['destination_server'] = Server.query.get_or_raise(kwargs.get('dst_server_id'))
        super().from_json(kwargs)
        try:
            o = cls.query.get((kwargs['file_id'], kwargs['dst_server_id']))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)


class File(UUIDistributedEntityMixin, SoftDeleteMixin, db.Model):
    __tablename__ = 'D_file'
    order = 20

    src_server_id = db.Column(typos.UUID, db.ForeignKey('D_server.id'), nullable=False)
    target = db.Column(db.Text, nullable=False)
    dest_folder = db.Column(db.Text)
    _old_target = db.Column("$$target", db.Text)
    l_mtime = db.Column(db.INTEGER)

    source_server = db.relationship("Server")
    destinations: t.List[FileServerAssociation] = db.relationship("FileServerAssociation", lazy='joined')

    __table_args__ = (db.UniqueConstraint('src_server_id', 'target'),)

    query_class = QueryWithSoftDelete

    def __init__(self, source_server: Server, target: str,
                 dest_folder=None, destination_servers: Destination_Servers = None, **kwargs):
        super().__init__(**kwargs)

        self.source_server = source_server
        self.target = target
        self.dest_folder = dest_folder
        dest = []
        for ds in destination_servers or []:
            if isinstance(ds, Server):
                dest.append(FileServerAssociation(file=self, destination_server=ds))
            elif isinstance(ds, dict):

                dest.append(
                    FileServerAssociation(file=self, destination_server=Server.query.get(ds.get('dst_server_id')),
                                          dest_folder=ds.get('dest_folder')))
            elif isinstance(ds, tuple):
                if isinstance(ds[0], Server):
                    dest.append(FileServerAssociation(file=self, destination_server=ds[0], dest_folder=ds[1]))
                else:
                    dest.append(
                        FileServerAssociation(file=self, destination_server=Server.query.get(ds[0]), dest_folder=ds[1]))
        self.destinations = dest

    def __str__(self):
        return f"{self.source_server}:{self.target}"

    def to_json(self, human=False, destinations=False, include: t.List[str] = None,
                exclude: t.List[str] = None, **kwargs):
        data = super().to_json(**kwargs)
        if self.source_server.id is None:
            raise RuntimeError('Set ids for servers before')
        data.update(target=self.target, dest_folder=self.dest_folder)
        if human:
            data.update(src_server=str(self.source_server.name))
            if destinations:
                dest = []
                for d in self.destinations:
                    dest.append(dict(dst_server=d.destination_server.name, dest_folder=d.dest_folder))
                data.update(destinations=dest)
        else:
            data.update(src_server_id=str(self.source_server.id))
            if destinations:
                dest = []
                for d in self.destinations:
                    dest.append(dict(dst_server_id=d.destination_server.id, dest_folder=d.dest_folder))
                data.update(destinations=dest)

        if include:
            data = {k: v for k, v in data.items() if k in include}
        if exclude:
            data = {k: v for k, v in data.items() if k not in exclude}
        return data

    @classmethod
    def from_json(cls, kwargs) -> 'File':
        kwargs = copy.deepcopy(kwargs)
        kwargs['source_server'] = Server.query.get_or_raise(kwargs.pop('src_server_id'))
        return super().from_json(kwargs)

    def delete(self):
        for d in self.destinations:
            d.delete()
        return super().delete()



import copy
from enum import Enum

from sqlalchemy import orm

from dimensigon import defaults
from dimensigon.domain.entities.base import UUIDistributedEntityMixin
from dimensigon.utils import typos
from dimensigon.utils.helpers import is_iterable_not_string
from dimensigon.utils.typos import MultiLine
from dimensigon.web import db


class ActionType(Enum):
    TEST = 0
    ANSIBLE = 1
    PYTHON = 2
    SHELL = 3
    ORCHESTRATION = 4
    REQUEST = 5
    NATIVE = 6


class ActionTemplate(UUIDistributedEntityMixin, db.Model):
    SEND_SOFTWARE = '00000000-0000-0000-000a-000000000001'
    WAIT_SERVERS = '00000000-0000-0000-000a-000000000002'
    ORCHESTRATION = '00000000-0000-0000-000a-000000000003'
    WAIT_ROUTE2SERVERS = '00000000-0000-0000-000a-000000000004'
    DELETE_SERVERS = '00000000-0000-0000-000a-000000000005'

    __tablename__ = 'D_action_template'
    order = 10

    name = db.Column(db.String(40), nullable=False)
    version = db.Column(db.Integer, nullable=False)
    action_type = db.Column(typos.Enum(ActionType), nullable=False)
    code = db.Column(db.Text)
    expected_stdout = db.Column(db.Text)
    expected_stderr = db.Column(db.Text)
    expected_rc = db.Column(db.Integer)
    system_kwargs = db.Column(db.JSON)
    pre_process = db.Column(db.Text)
    post_process = db.Column(db.Text)
    schema = db.Column(db.JSON)
    description = db.Column(db.Text)

    def __init__(self, name: str, version: int, action_type: ActionType, code: MultiLine = None,
                 expected_stdout: MultiLine = None, expected_stderr: MultiLine = None,
                 expected_rc: int = None, system_kwargs: typos.Kwargs = None, pre_process: MultiLine = None,
                 post_process: MultiLine = None, schema: typos.Kwargs = None, description: MultiLine = None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.version = version
        self.action_type = action_type
        self.code = '\n'.join(code) if is_iterable_not_string(code) else code
        self.schema = schema or {}
        self.expected_stdout = '\n'.join(expected_stdout) if is_iterable_not_string(
            expected_stdout) else expected_stdout
        self.expected_stderr = '\n'.join(expected_stderr) if is_iterable_not_string(
            expected_stderr) else expected_stderr
        self.expected_rc = expected_rc
        self.system_kwargs = system_kwargs or {}
        self.pre_process = '\n'.join(pre_process) if is_iterable_not_string(pre_process) else pre_process
        self.post_process = '\n'.join(post_process) if is_iterable_not_string(post_process) else post_process
        self.description = '\n'.join(description) if is_iterable_not_string(description) else description

    __table_args__ = (db.UniqueConstraint('name', 'version'),)

    @orm.reconstructor
    def init_on_load(self):
        self.system_kwargs = self.system_kwargs or {}

    def __str__(self):
        return f"{self.name}.ver{self.version}"

    def to_json(self, split_lines=False, **kwargs):
        data = super().to_json(**kwargs)
        data.update(name=self.name, version=self.version,
                    action_type=self.action_type.name)
        if self.code is not None:
            data.update(code=self.code.split('\n') if split_lines else self.code)
        if self.schema:
            data.update(schema=self.schema)
        if self.system_kwargs:
            data.update(system_kwargs=self.system_kwargs)
        if self.expected_stdout is not None:
            data.update(expected_stdout=self.expected_stdout.split('\n') if split_lines else self.expected_stdout)
        if self.expected_stderr is not None:
            data.update(expected_stderr=self.expected_stderr.split('\n') if split_lines else self.expected_stderr)
        if self.expected_rc is not None:
            data.update(expected_rc=self.expected_rc)
        if self.post_process is not None:
            data.update(post_process=self.post_process.split('\n') if split_lines else self.post_process)
        if self.pre_process is not None:
            data.update(pre_process=self.pre_process.split('\n') if split_lines else self.pre_process)
        return data

    @classmethod
    def from_json(cls, kwargs):
        kwargs = copy.deepcopy(kwargs)
        kwargs['action_type'] = ActionType[kwargs.get('action_type')]
        return super().from_json(kwargs)

    @classmethod
    def set_initial(cls, session=None):
        from dimensigon.domain.entities import bypass_datamark_update

        if session is None:
            session = db.session

        with bypass_datamark_update(session):
            at = session.query(cls).get(cls.SEND_SOFTWARE)
            if at is None:
                at = ActionTemplate(name='send software', version=1, action_type=ActionType.NATIVE,
                                    expected_rc=201, last_modified_at=defaults.INITIAL_DATEMARK,
                                    schema={"input": {"software": {"type": "string",
                                                                   "description": "software name or ID to send. If "
                                                                                  "name specified and version not set, "
                                                                                  "biggest version will be taken"},
                                                      "version": {"type": "string",
                                                                  "description": "software version to take"},
                                                      "server": {"type": "string",
                                                                 "description": "destination server id"},
                                                      "dest_path": {"type": "string",
                                                                    "description": "destination path to send software"},
                                                      "chunk_size": {"type": "integer"},
                                                      "max_senders": {"type": "integer"},
                                                      },
                                            "required": ["software", "server"],
                                            "output": ["file"]
                                            },
                                    id=cls.SEND_SOFTWARE,
                                    post_process="import json\nif cp.success:\n  json_data=json.loads(cp.stdout)\n  vc.set('file', "
                                                 "json_data.get('file'))")

                session.add(at)
            at = session.query(cls).get(cls.WAIT_SERVERS)
            if at is None:
                at = ActionTemplate(name='wait servers', version=1, action_type=ActionType.NATIVE,
                                    description="waits server_names to join to the dimension",
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    schema={"input": {"server_names": {"type": ["array", "string"],
                                                                       "items": {"type": "string"}},
                                                      },
                                            "required": ["server_names"]
                                            },
                                    id=cls.WAIT_SERVERS)
                session.add(at)
            at = session.query(cls).get(cls.ORCHESTRATION)
            if at is None:
                at = ActionTemplate(name='orchestration', version=1, action_type=ActionType.ORCHESTRATION,
                                    description="launches an orchestration",
                                    schema={"input": {"orchestration": {"type": "string",
                                                                        "description": "orchestration name or ID to "
                                                                                       "execute. If no version "
                                                                                       "specified, the last one will "
                                                                                       "be executed"},
                                                      "version": {"type": "integer"},
                                                      "hosts": {"type": ["string", "array", "object"],
                                                                "items": {"type": "string"},
                                                                "minItems": 1,
                                                                "patternProperties": {
                                                                    ".*": {"anyOf": [{"type": "string"},
                                                                                     {"type": "array",
                                                                                      "items": {"type": "string"},
                                                                                      "minItems": 1
                                                                                      },
                                                                                     ]
                                                                           },
                                                                },
                                                                },
                                                      },
                                            "required": ["orchestration", "hosts"]
                                            },
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    id=cls.ORCHESTRATION)
                session.add(at)
            at = session.query(cls).get(cls.WAIT_ROUTE2SERVERS)
            if at is None:
                at = ActionTemplate(name='wait route to servers', version=1, action_type=ActionType.NATIVE,
                                    description="waits until we have a valid route to a server",
                                    schema={"input": {"server_names": {"type": ["array", "string"],
                                                                       "items": {"type": "string"}},
                                                      },
                                            "required": ["server_names"]
                                            },
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    id=cls.WAIT_ROUTE2SERVERS)
                session.add(at)
            at = session.query(cls).get(cls.DELETE_SERVERS)
            if at is None:
                at = ActionTemplate(name='delete servers', version=1, action_type=ActionType.NATIVE,
                                    description="deletes server_names from the dimension",
                                    schema={"input": {"server_names": {"type": ["array", "string"],
                                                                       "items": {"type": "string"}},
                                                      },
                                            "required": ["server_names"]
                                            },
                                    last_modified_at=defaults.INITIAL_DATEMARK,
                                    id=cls.DELETE_SERVERS)
                session.add(at)



__author__ = "Joan Prat"
__copyright__ = "Copyright 2019, The Dimensigon project"
__credits__ = ["Joan Prat", "Daniel Moya"]
__license__ = "GNU General Public License v3 or later (GPLv3+)"
__version__ = "0.3.5"
__maintainer__ = "Joan Prat"
__email__ = "joan.prat@dimensigon.com"
__status__ = "development"



import logging
import multiprocessing as mp
import os
import time
import typing as t

from gunicorn.app.base import Application
from gunicorn.pidfile import Pidfile

from dimensigon import defaults
from dimensigon.exceptions import DimensigonError
from dimensigon.use_cases.base import TerminateInterrupt
from dimensigon.use_cases.catalog import CatalogManager
from dimensigon.use_cases.cluster import ClusterManager
from dimensigon.use_cases.file_sync import FileSync
# from dimensigon.use_cases.log_sender import LogSender
from dimensigon.use_cases.mptools import MainContext
from dimensigon.use_cases.mptools_events import EventMessage
from dimensigon.use_cases.routing import RouteManager
from dimensigon.utils.typos import Id
from dimensigon.web import DimensigonFlask, create_app, threading

_logger = logging.getLogger("dm")


def _sleep_secs(max_sleep, end_time=999999999999999.9):
    # Calculate time left to sleep, no less than 0
    return max(0.0, min(end_time - time.time(), max_sleep))


class GunicornApp(Application):
    dm = None

    def __init__(self, application, options=None):
        """ Construct the Application. Default gUnicorn configuration is loaded """

        self.application = application
        self.options = options or {}
        self.process = None

        # if port, or host isn't set-- run from os.environments
        #
        super(GunicornApp, self).__init__()

    def init(self, parser, opts, args):
        pass

    def load_config(self):
        # Load up the any app specific configuration
        for k, v in self.options.items():
            self.cfg.set(k.lower(), v)

    def load(self):
        return self.application


class Dimensigon:
    def __init__(self):
        self.flask_app: t.Optional[DimensigonFlask] = None
        self.gunicorn: t.Optional[GunicornApp] = None
        self.server: t.Optional[mp.Process] = None
        self.config = Config(self)

        # processes
        self.manager = mp.Manager()  # shared memory between processes
        self.cluster_manager: t.Optional[ClusterManager] = None
        self.file_sync: t.Optional[FileSync] = None
        self.route_manager: t.Optional[RouteManager] = None
        self.catalog_manager: t.Optional[CatalogManager] = None

        self.STOP_WAIT_SECS = 90
        self.engine = None  # set on setup_dm function
        self.get_session = None  # set on setup_dm function
        self._main_ctx = MainContext()
        self.server_id: t.Optional[Id] = None
        self.pid = None
        self.pidfile = None

    def create_flask_instance(self):
        if self.flask_app is None:
            self.flask_app = create_app(self.config.flask_conf)
            self.flask_app.dm = self

    def create_gunicorn_instance(self):
        if self.gunicorn is None:
            self.gunicorn = GunicornApp(self.flask_app, self.config.http_conf)
            self.gunicorn.dm = self

    def set_catalog_manager(self):
        if self.catalog_manager is None:
            self.catalog_manager = CatalogManager(None, None, None, None, None, self)

    def create_processes(self):
        self.cluster_manager = self._main_ctx.Proc(ClusterManager, self)
        self.cluster_manager.SHUTDOWN_WAIT_SECS = 90

        self.route_manager = self._main_ctx.Proc(RouteManager, self)
        self.file_sync = self._main_ctx.Proc(FileSync, self)
        self.catalog_manager = self._main_ctx.Thread(CatalogManager, self)
        # self.log_sender = LogSender(self)  # log sender embedded in file_sync process
        if self.config.flask:
            self.http_server = mp.Process(target=self.flask_app.run, name="Flask server",
                                          kwargs=dict(host='0.0.0.0', port=defaults.DEFAULT_PORT, ssl_context='adhoc'))
        else:
            self.http_server = mp.Process(target=self.gunicorn.run, name="Gunicorn server")

    def bootstrap(self):
        self.create_flask_instance()
        self.create_gunicorn_instance()
        self.create_processes()

    def make_first_request(self):
        from dimensigon.domain.entities import Server
        import dimensigon.web.network as ntwrk

        with self.flask_app.app_context():
            start = time.time()
            while True:
                resp = ntwrk.get(Server.get_current(), 'root.home', timeout=1)
                if not resp.ok and time.time() - start < 30:
                    time.sleep(0.5)
                else:
                    break
            self._main_ctx.publish_q.safe_put(EventMessage("Listening", source="Dimensigon"))

    def start_server(self):
        # if hasattr(self.cluster_manager, 'notify_cluster'):
        #     self.flask_app.before_first_request(self.cluster_manager.notify_cluster)
        th = threading.Timer(interval=4, function=self.make_first_request)
        th.start()
        self.http_server.start()

    def start(self):
        """starts dimensigon server"""
        _logger.info(f"Starting Dimensigon ({os.getpid()})")
        self._main_ctx.init_signals()
        self.pid = os.getpid()
        pidname = self.config.pidfile
        self.pidfile = Pidfile(pidname)
        self.pidfile.create(self.pid)
        self.bootstrap()
        self.flask_app.bootstrap()

        # self.cluster_manager.start()
        # self.file_sync.start()
        # self.log_sender.start() # log sender embedded in file_sync process
        self.start_server()
        try:
            while not self._main_ctx.shutdown_event.is_set():
                self._main_ctx.forward_events()
        except (TerminateInterrupt, KeyboardInterrupt):
            pass
        self.shutdown()

    def shutdown(self):
        _logger.info(f"Shutting down Dimensigon")
        self.http_server.terminate()
        self.http_server.terminate()
        self.http_server.join(90)
        if self.http_server.is_alive():
            self.http_server.kill()
        self._main_ctx.stop()


class Config:

    def __init__(self, dm: Dimensigon):
        self.dm = dm

        # Directory that holds the configuration
        self.config_dir: t.Optional[str] = None

        # If set, process should upgrade as soon as a neighbour has a higher version
        self.auto_upgrade: bool = True

        # sets a security layer on top of HTTP
        self.security_layer: bool = True

        # allow pass through security layer without encrypt packet with header D-Securizer: plain
        self.security_layer_antidote: bool = False

        # pidfile name
        self.pidfile: str = None

        # runs the scheduler
        self.scheduler: bool = True

        # database uri
        self.db_uri: t.Optional[str] = None

        # http configuration
        self.http_conf = {}

        # flask configuration
        self.flask_conf = {}

        # Run configuration (used for elevator to load same configuration)
        self.run_config = {}

        self.debug: bool = False

        self.flask: bool = False

        # forces the process to scan on startup
        self.force_scan: bool = False

        # Run route table, catalog and cluster refresh every minutes
        # self.refresh_interval: dt.timedelta = defaults.

    def path(self, *path: str) -> str:
        """Generate path to the file within the configuration directory.
        """
        if self.config_dir is None:
            raise DimensigonError("config_dir is not set")
        return os.path.join(self.config_dir, *path)
